"""Cell list to facilitate neighbor searching."""

import logging

_LOGGER = logging.getLogger(__name__)


class Cells:
    """Accelerate the search for nearby atoms.

    A pure all versus all search is O(n^2) - for every atom, every other atom
    must be searched. This is rather inefficient, especially for large
    biomolecules where cells may be tens of angstroms apart. The cell class
    breaks down the xyz biomolecule space into several 3-D cells of desired
    size - then by simply examining atoms that fall into the adjacent cells
    one can quickly find nearby cells.
    """

    def __init__(self, cellsize):
        """Initialize the cells.

        :param cellsize:  the size of each cell (in Angstroms)
        :type cellsize:  int
        """
        self.cellmap = {}
        self.cellsize = cellsize

    def assign_cells(self, biomolecule):
        """Place each atom in a virtual cell for easy neighbor comparison.

        :param biomolecule:  biomolecule with atoms to assign to cells
        :type biomolecule:  Biomolecule
        """
        for atom in biomolecule.atoms:
            atom.cell = None
            self.add_cell(atom)

    def add_cell(self, atom):
        """Add an atom to the cell.

        :param atom:  the atom to add
        :type atom:  Atom
        """
        size = self.cellsize
        x = atom.x
        x = (int(x) - 1) // size * size if x < 0 else int(x) // size * size
        y = atom.y
        y = (int(y) - 1) // size * size if y < 0 else int(y) // size * size
        z = atom.z
        z = (int(z) - 1) // size * size if z < 0 else int(z) // size * size
        key = (x, y, z)
        try:
            self.cellmap[key].append(atom)
        except KeyError:
            self.cellmap[key] = [atom]
        atom.cell = key

    def remove_cell(self, atom):
        """Remove an atom from a cell.

        :param atom:  the atom to remove
        :type atom:  Atom
        """
        oldcell = atom.cell
        if oldcell is None:
            return
        atom.cell = None
        self.cellmap[oldcell].remove(atom)

    def get_near_cells(self, atom):
        """Find all atoms in cells bordering an atom.

        :param atom:  the atom to test
        :type atom:  Atom
        :return:  a list of nearby atoms
        :rtype:  [Atom]
        """
        closeatoms = []
        cell = atom.cell
        if cell is not None:
            x = cell[0]
            y = cell[1]
            z = cell[2]
            size = self.cellsize
            for i in range(-1 * size, 2 * size, size):
                for j in range(-1 * size, 2 * size, size):
                    for k in range(-1 * size, 2 * size, size):
                        newkey = (x + i, y + j, z + k)
                        try:
                            newatoms = self.cellmap[newkey]
                            for atom2 in newatoms:
                                if atom == atom2:
                                    continue
                                closeatoms.append(atom2)
                        except KeyError:
                            pass

        return closeatoms



"""Force fields are fun.

The forcefield structure is modeled off of the structures.py file, where each
forcefield is considered a chain of residues of atoms.

.. codeauthor::  Todd Dolinsky
.. codeauthor::  Yong Huang
"""

import logging
import re
from xml import sax

from . import io

_LOGGER = logging.getLogger(__name__)


class ForcefieldHandler(sax.ContentHandler):
    """Process XML-format force field parameter files."""

    def __init__(self, map_, reference):
        """Initialize handler

        :param map_:  dictionary of paramaeter information
        :type map_:  dict
        :param reference:  reference map for force field
        :type reference:  dict
        """
        self.oldresname = None
        self.oldatomname = None
        self.curelement = None
        self.newatomname = None
        self.newresname = None
        self.atommap = {}
        self.map = map_
        self.reference = reference

    @classmethod
    def update_map(cls, toname, fromname, map_):
        """Update the given map by adding a pointer from a new name to an object.

        .. todo:: Should this be a staticmethod instead of classmethod?

        :param toname:  the new name for the object
        :type toname:  str
        :param fromname:  the old name for the object
        :type fromname:  str
        :param map_:  a dictionary of forcefield-related items
        :type map_: dict
        """
        fromobj = map_[fromname]
        if isinstance(fromobj, ForcefieldResidue):
            if toname not in map_:
                newres = ForcefieldResidue(fromname)
                map_[toname] = newres
            for atomname in fromobj.atoms:
                map_[toname].atoms[atomname] = fromobj.atoms[atomname]
        elif isinstance(fromobj, ForcefieldAtom):
            map_[toname] = fromobj

    @classmethod
    def find_matching_names(cls, regname, map_):
        """Find strings in the map that match the given regular expression.

        .. todo:: Should this be a staticmethod instead of classmethod?

        :param regname: the regular expression to search for in the map
        :type regname:  str
        :param map_:  the dictionary to search
        :type map_: dict
        :return:  a list of regular expression match objects for dictionary
            keys that match the regular expression.
        :rtype:  [re.Match]
        """
        name_list = []
        regname += "$"
        # Find the existing items that match this string
        for name in map_:
            regexp = re.compile(regname).match(name)
            if regexp:
                name_list.append(regexp)
        return name_list

    def startElement(self, name, _):
        """Start XML element parsing.

        :param name:  element name
        :type name:  str
        """
        if name != "name":
            self.curelement = name

    def endElement(self, name):
        """End XML element parsing.

        :param name:  the name of the element
        :type name:  str
        """
        if name == "residue":
            if self.oldresname is not None:  # Make a new residue hook
                newreslist = self.find_matching_names(
                    self.newresname, self.reference
                )
                # Multiple new residues
                if self.oldresname.find("$group") >= 0:
                    for resitem in newreslist:
                        resname = resitem.string
                        group = resitem.group(1)
                        fromname = self.oldresname.replace("$group", group)
                        if fromname in self.map:
                            self.update_map(resname, fromname, self.map)
                else:  # Work with a single new residue name
                    _ = self.find_matching_names(self.oldresname, self.map)
                    for resitem in newreslist:
                        resname = resitem.string
                        self.update_map(resname, self.oldresname, self.map)
            # If this was only a residue conversion, exit
            if self.atommap == {}:
                self.oldresname = None
                self.newresname = None
                return None
            # Apply atom conversions for all appropriate residues
            resmatchlist = self.find_matching_names(self.newresname, self.map)
            for resitem in resmatchlist:
                residue = self.map[resitem.string]
                for newname in self.atommap:
                    oldname = self.atommap[newname]
                    if oldname not in residue.atoms:
                        continue
                    self.update_map(newname, oldname, residue.atoms)
            # Clean up
            self.oldresname = None
            self.newresname = None
            self.atommap = {}
        elif name == "atom":
            self.atommap[self.newatomname] = self.oldatomname
            self.oldatomname = None
            self.newatomname = None
        else:  # Just free the current element namespace
            self.curelement = ""
        return self.map

    def characters(self, text):
        """Parse text information within XML tag.

        :param text:  the text value between the XML tags
        :type test:  str
        """
        if text.isspace():
            return
        text = str(text)
        if self.curelement == "residue":
            self.newresname = text
        elif self.curelement == "atom":
            self.newatomname = text
        elif self.curelement == "useatomname":
            self.oldatomname = text
        elif self.curelement == "useresname":
            self.oldresname = text


class Forcefield:
    """Parameter definitions for a given forcefield.

    .. note::
        Pass ff and ff names file like objects rather than sorting out whether
        to use user-created files here.

    The forcefield class contains definitions for a given forcefield.
    Each forcefield object contains a dictionary of residues, with each residue
    containing a dictionary of atoms.
    Dictionaries are used instead of lists as the ordering is not important.
    The forcefield definition files are unedited, directly from the
    forcefield - all transformations are done within.
    """

    def __init__(self, ff_name, definition, userff, usernames=None):
        """Initialize the class by parsing the definition file.

        .. todo:: Why are files being loaded so deep in this function?

        :param ff_name: the name of the forcefield (can be None)
        :type ff_name:  str
        :param definition:  the definition object for the Forcefield
        :type definition:  Definition
        :param userff:  path for user-defined forcefields file
        :type userff:  str
        :param usernames:  path to user-defined atom/residue names file
        :type usernames:  str
        :raises ValueError:  if invalid force field names specified
        """
        self.map = {}
        self.name = str(ff_name)
        defpath = ""
        defpath = io.test_dat_file(ff_name) if userff is None else userff
        with open(defpath, encoding="utf-8") as ff_file:
            lines = ff_file.readlines()
            for line in lines:
                if not line.startswith("#"):
                    fields = line.split()
                    if fields == []:
                        continue
                    try:
                        resname = fields[0]
                        atomname = fields[1]
                        charge = float(fields[2])
                        radius = float(fields[3])
                    except ValueError:
                        txt = (
                            "Unable to recognize user-defined forcefield file"
                        )
                        txt += f" {defpath}!" if defpath != "" else "!"
                        txt += " Please use a valid parameter file."
                        raise ValueError(txt)
                    try:
                        group = fields[4]
                        atom = ForcefieldAtom(
                            atomname, charge, radius, resname, group
                        )
                    except IndexError:
                        atom = ForcefieldAtom(
                            atomname, charge, radius, resname
                        )

                    my_residue = self.get_residue(resname)
                    if my_residue is None:
                        my_residue = ForcefieldResidue(resname)
                        self.map[resname] = my_residue
                    my_residue.add_atom(atom)
        # Now parse the XML file, associating with FF objects -
        # This is not necessary (if canonical names match ff names)
        try:
            defpath = io.test_names_file(ff_name)
        except FileNotFoundError:
            defpath = None
        if usernames:
            names_path = usernames
        elif defpath:
            names_path = defpath
        else:
            raise ValueError("Unable to identify .names file.")
        handler = ForcefieldHandler(self.map, definition.map)
        sax.make_parser()
        with open(names_path, encoding="utf-8") as namesfile:
            sax.parseString(namesfile.read(), handler)

    def has_residue(self, resname):
        """Check if the residue name is in the map or not.

        :param resname:  the residue name to search
        :type resname:  str
        :return:  indication of whether resname is in map
        :rtype:  bool
        """
        return resname in self.map

    def get_residue(self, resname):
        """Return the residue object with the given resname.

        :param resname:  the name of the residue
        :type resname:  str
        :return:  residue object
        :rtype:  ForcefieldResidue
        """
        if self.has_residue(resname):
            return self.map[resname]
        return None

    def get_names(self, resname, atomname):
        """Get forcefield names associated with residue and atom.

        The names passed in point to ForcefieldResidue and ForcefieldAtom
        objects which may have different names; grab these names and return.

        :param resname:  the residue name
        :type resname:  str
        :param atomname:  the atom name
        :type atomname:  str
        :return:  (forcefield's name for this residue, forcefield's name for
            this atom)
        :rtype:  (str, str)
        """
        rname = None
        aname = None
        if resname in self.map:
            res = self.map[resname]
            if res.has_atom(atomname):
                atom = res.atoms[atomname]
                aname = atom.name
                rname = atom.resname
        return rname, aname

    def get_group(self, resname, atomname):
        """Get the group/type associated with the input fields.

        :param resname:  the residue name
        :type resname:  str
        :param atomname:  the atom name
        :type atomname:  str
        :return:  group name or empty string
        :rtype:  str
        """
        group = ""
        if resname in self.map:
            resid = self.map[resname]
            if resid.has_atom(atomname):
                atom = resid.atoms[atomname]
                group = atom.group
        return group

    def get_params(self, resname, atomname):
        """Get the charge and radius parameters for an atom in a residue.

        The residue itself is needed instead of simply its name because the
        forcefield may use a different residue name than the standard amino
        acid name.

        .. todo::
           Why do both :func:`get_params` and :func:`get_params1` exist?

        :param resname:  the residue name
        :type resname:  str
        :param atomname:  the atom name
        :type atomname:  str
        :return:  (charge of the atom, radius of the atom)
        :rtype:  (float, float)
        """
        charge = None
        radius = None
        if resname in self.map:
            resid = self.map[resname]
            if resid.has_atom(atomname):
                atom = resid.atoms[atomname]
                charge = atom.charge
                radius = atom.radius
        return charge, radius

    def get_params1(self, residue, name):
        """Get the charge and radius parameters for an atom in a residue.

        The residue itself is needed instead of simply its name because the
        forcefield may use a different residue name than the standard amino
        acid name.

        .. todo::
           Why do both :func:`get_params` and :func:`get_params1` exist?

        :param resname:  the residue name
        :type resname:  str
        :param name:  the atom name
        :type name:  str
        :return:  (charge of the atom, radius of the atom)
        :rtype:  (float, float)
        """
        charge = None
        radius = None
        resname = ""
        atomname = ""
        if self.name == "amber":
            resname, atomname = self.get_amber_params(residue, name)
        elif self.name == "charmm":
            resname, atomname = self.get_charmm_params(residue, name)
        elif self.name == "parse":
            resname, atomname = self.get_parse_params(residue, name)
        else:
            resname = residue.name
            atomname = name
        defresidue = self.get_residue(resname)
        if defresidue is None:
            return charge, radius
        atom = defresidue.get_atom(atomname)
        if atom is not None:
            charge = atom.charge
            radius = atom.radius
        return charge, radius

    @classmethod
    def get_amber_params(cls, residue, name):
        """Get AMBER forcefield definitions.

        .. todo::  Should this be a staticmethod or a classmethod?

        .. todo::
           Figure out why :makevar:`residue.type` has :class:`int` values

        :param residue:  the residue
        :type residue:  Residue
        :param name:  the atom name
        :type name:  str
        :return:  (forcefield name of residue, forcefield name of atom)
        :rtype:  (str, str)
        """
        atomname = name
        res_type = residue.type
        resname = residue.naname if res_type == 4 else residue.name
        # Residue Substitutions
        if residue.name == "CYS" and "HG" not in residue.map:
            resname = "CYX"
        elif residue.name == "HIS":
            if "HD1" in residue.map and "HE2" in residue.map:
                resname = "HIP"
            elif "HD1" in residue.map:
                resname = "HID"
            elif "HE2" in residue.map:
                resname = "HIE"
            else:
                resname = "HID"  # Default for no hydrogens
        elif residue.name == "HSP":
            resname = "HIP"
        elif residue.name == "HSE":
            resname = "HIE"
        elif residue.name == "HSD":
            resname = "HID"
        elif residue.name in ["GLU", "GLH"]:
            if "HE1" in residue.map:
                resname = "GLH"
                if atomname == "HE1":
                    atomname = "HE2"
                elif atomname == "OE1":
                    atomname = "OE2"
                elif atomname == "OE2":
                    atomname = "OE1"
            elif "HE2" in residue.map:
                resname = "GLH"
        elif residue.name in ["ASP", "ASH"]:
            if "HD1" in residue.map:
                resname = "ASH"
                if atomname == "HD1":
                    atomname = "HD2"
                elif atomname == "OD1":
                    atomname = "OD2"
                elif atomname == "OD2":
                    atomname = "OD1"
            elif "HD2" in residue.map:
                resname = "ASH"
        if residue.is_c_term:
            resname = "C" + resname
        elif residue.is_n_term:
            resname = "N" + resname
        # Atom Substitutions
        if resname == "ILE":
            if atomname == "CD":
                atomname = "CD1"
        elif resname == "WAT":
            if atomname == "O":
                atomname = "OW"
            elif atomname in ["H1", "H2"]:
                atomname = "HW"
        # N-terminal
        if resname[0] == "N" and resname != "NME" and atomname == "H":
            atomname = "H1"
        if resname in ["CCYS", "NCYS"] and atomname == "HG":
            atomname = "HSG"
        if resname == "CYM" and atomname == "H":
            atomname = "HN"
        if residue.is_n_term and resname == "NPRO" and atomname == "HN2":
            atomname = "H2"
        if residue.is_n_term and resname == "NPRO" and atomname == "HN1":
            atomname = "H3"
        return resname, atomname

    @classmethod
    def get_parse_params(cls, residue, name):
        """Get PARSE forcefield definitions.

        .. todo::  Should this be a staticmethod or a classmethod?

        :param residue:  the residue
        :type residue:  Residue
        :param name:  the atom name
        :type name:  str
        :return:  (forcefield name of residue, forcefield name of atom)
        :rtype:  (str, str)
        """
        # TODO - this is a crazy list of conditionals!
        atomname = name
        resname = residue.name
        # Terminal/Water Substitutions
        nterm = residue.is_n_term
        cterm = residue.is_c_term
        if nterm and resname != "ACE":
            if resname == "PRO" and nterm == 2:
                resname = "PR+"
                if atomname == "H2":
                    atomname = "HN1"
                elif atomname == "H3":
                    atomname = "HN2"
            elif resname == "PRO" and nterm == 1:
                resname = "PRN"
                if atomname in ["H2", "H3"]:
                    atomname = "HN"
            elif nterm == 2:  # Neutral
                # TODO - there are a lot of hard-coded repeated lists like this
                # They should be replaced with module-level variables.
                if atomname in ["N", "H", "H2", "H3", "CA", "HA", "C", "O"]:
                    resname = "BKN"
                if atomname == "H":
                    atomname = "H1"
                if atomname == "H3":
                    atomname = "H2"
            elif nterm == 3:  # Positive
                if atomname in ["N", "H", "H2", "H3", "CA", "HA", "C", "O"]:
                    resname = "BK+"
                if atomname == "H":
                    atomname = "H1"
        elif cterm:
            if atomname == "O":
                atomname = "O1"
            elif atomname == "OXT":
                atomname = "O2"
            if cterm == 1 and atomname in [
                "N",
                "H",
                "HA",
                "CA",
                "C",
                "O1",
                "O2",
            ]:
                resname = "BK-"
            elif cterm == 2 and atomname in [
                "N",
                "H",
                "HA",
                "CA",
                "C",
                "O1",
                "O2",
                "HO",
            ]:
                if atomname == "HO":
                    atomname = "H2"
                resname = "BKC"
            # print 'Cterm resname is',resname
        elif residue.type == 3:
            resname = "H2O"
            if atomname == "O":
                atomname = "OH"
            elif atomname == "H1":
                atomname = "HH1"
            elif atomname == "H2":
                atomname = "HH2"
        # Residue Substitutions
        if resname == "HSD":
            resname = "HID"
        elif resname in ["HIE", "HSE"]:
            resname = "HIS"
        elif resname in ["HIP", "HSP"]:
            resname = "HI+"
        elif resname == "ILE":
            if atomname == "HG12":
                atomname = "HG11"
            elif atomname == "HG13":
                atomname = "HG12"
            elif atomname == "CD":
                atomname = "CD1"
        elif resname == "CYS" and "HG" not in residue.map:
            resname = "CSS"
        # Histidine
        elif resname == "HIS":
            if "HD1" in residue.map and "HE2" in residue.map:
                resname = "HI+"
            elif "HD1" in residue.map:
                resname = "HID"
            elif "HE2" in residue.map:
                resname = "HIS"
        elif resname in ["GLU", "GLH"]:
            if "HE1" in residue.map:
                resname = "GL0"
                if atomname == "HE1":
                    atomname = "HE2"
                elif atomname == "OE1":
                    atomname = "OE2"
                elif atomname == "OE2":
                    atomname = "OE1"
            elif "HE2" in residue.map:
                resname = "GL0"
        elif resname in ["ASP", "ASH"]:
            if "HD1" in residue.map:
                resname = "AS0"
                if atomname == "HD1":
                    atomname = "HD2"
                elif atomname == "OD1":
                    atomname = "OD2"
                elif atomname == "OD2":
                    atomname = "OD1"
            elif "HD2" in residue.map:
                resname = "AS0"
        elif resname == "ACE":
            if atomname == "HH31":
                atomname = "HA1"
            elif atomname == "HH32":
                atomname = "HA2"
            elif atomname == "HH33":
                atomname = "HA3"
            elif atomname == "CH3":
                atomname = "CA"
        elif resname == "TYR":
            if "HH" not in residue.map:
                resname = "TYM"
        elif resname == "TYM":
            resname = "TY-"
        elif resname == "CYM":
            resname = "CY-"
        elif resname == "LYN":
            resname = "LY0"
        # Neutral LYS and neutral ARG detection based on hydrogens - added by
        # Jens
        elif resname == "LYS":
            if "HZ3" not in residue.map:
                resname = "LY0"
        elif resname == "ARG":
            if "HE" not in residue.map:
                resname = "AR0"
        elif resname == "NME":
            resname = "N-M"
            if atomname == "CH3":
                atomname = "CA"
            elif atomname == "H":
                atomname = "H1"
            elif atomname.startswith("HH"):
                atomname = "HA" + atomname[-1]
        # Hydrogen Substitutions
        if atomname == "H":
            atomname = "HN"
        elif atomname == "HA2":
            atomname = "HA1"
        elif atomname == "HA3":
            atomname = "HA2"
        elif atomname == "HB2" and resname not in ["ALA"]:
            atomname = "HB1"
        elif atomname == "HB3" and resname not in ["ALA"]:
            atomname = "HB2"
        elif atomname == "HD2" and resname not in ["HIS", "HI+", "HID", "AS0"]:
            atomname = "HD1"
        elif atomname == "HD3" and resname not in ["HIS", "HI+", "HID"]:
            atomname = "HD2"
        elif atomname == "HE2" and resname not in [
            "TRP",
            "HIS",
            "HI+",
            "HID",
            "GL0",
        ]:
            atomname = "HE1"
        elif atomname == "HE3" and resname not in ["TRP", "HIS", "HI+", "HID"]:
            atomname = "HE2"
        elif atomname == "HG2":
            atomname = "HG1"
        elif atomname == "HG3":
            atomname = "HG2"
        elif atomname == "HZ2" and resname == "LY0":
            atomname = "HZ1"
        elif atomname == "HZ3" and resname == "LY0":
            atomname = "HZ2"
        return resname, atomname

    @classmethod
    def get_charmm_params(cls, residue, name):
        """Get CHARMM forcefield definitions.

        .. todo::  Should this be a staticmethod or a classmethod?

        .. todo::
           Figure out why :makevar:`residue.type` has :class:`int` values

        :param residue:  the residue
        :type residue:  Residue
        :param name:  the atom name
        :type name:  str
        :return:  (forcefield name of residue, forcefield name of atom)
        :rtype:  (str, str)
        """
        resname = residue.name
        atomname = name
        #  Nucleic Acid Substitutions
        if residue.type == 4:
            resname = resname[0]
            if resname == "A":
                resname = "ADE"
            elif resname == "C":
                resname = "CYT"
            elif resname == "G":
                resname = "GUA"
            elif resname == "T":
                resname = "THY"
                if atomname == "C7":
                    atomname = "C5M"
                elif atomname == "H71":
                    atomname = "H51"
                elif atomname == "H72":
                    atomname = "H52"
                elif atomname == "H73":
                    atomname = "H53"
            elif resname == "U":
                resname = "URA"
            if atomname == "H5'1":
                atomname = "H5'"
            elif atomname == "H5'2":
                atomname = "H5''"
            elif atomname == "H2'1":
                atomname = "H2'"
            elif atomname in ["H2'2", "HO'2"]:
                atomname = "H2''"
            if residue.get_atom("O2'") is None and atomname in [
                "C2'",
                "H2'",
                "H2''",
            ]:
                resname = "DEO1"
            if residue.get_atom("H5T") is not None and atomname in [
                "H5T",
                "O5'",
                "C5'",
            ]:
                resname = "5TER"
            if residue.get_atom("H3T") is not None and atomname in [
                "H3T",
                "O3'",
                "C3'",
            ]:
                resname = "3TER"
        # Terminal/Water Substitutions
        if residue.is_n_term:
            if resname == "GLY" and atomname in [
                "N",
                "H",
                "H2",
                "H3",
                "CA",
                "HA2",
                "HA3",
            ]:
                resname = "GLYP"
                if atomname == "H":
                    atomname = "HT1"
                elif atomname == "H2":
                    atomname = "HT2"
                elif atomname == "H3":
                    atomname = "HT3"
            elif resname == "PRO" and atomname in [
                "N",
                "HN1",
                "HN2",
                "CD",
                "CA",
                "HD1",
                "HD2",
                "HA",
                "H2",
                "H3",
            ]:
                resname = "PROP"
                if atomname == "H2":
                    atomname = "HN1"
                elif atomname == "H3":
                    atomname = "HN2"
            elif resname == "ACE":
                if atomname == "CH3":
                    atomname = "CAY"
                elif atomname == "HH31":
                    atomname = "HY1"
                elif atomname == "HH32":
                    atomname = "HY2"
                elif atomname == "HH33":
                    atomname = "HY3"
                elif atomname == "C":
                    atomname = "CY"
                elif atomname == "O":
                    atomname = "OY"
            else:
                if atomname in ["N", "H", "H2", "H3", "CA", "HA"]:
                    resname = "NTER"
                    if atomname == "H":
                        atomname = "HT1"
                    elif atomname == "H2":
                        atomname = "HT2"
                    elif atomname == "H3":
                        atomname = "HT3"
        elif residue.is_c_term:
            if atomname in ["O", "OXT", "C"]:
                resname = "CTER"
                if atomname == "O":
                    atomname = "OT1"
                elif atomname == "OXT":
                    atomname = "OT2"
        elif residue.type == 3:
            resname = "TP3M"
            if atomname == "O":
                atomname = "OH2"
        # Residue substitutions
        if resname == "ILE":
            if atomname == "CD1":
                atomname = "CD"
            elif atomname == "HD11":
                atomname = "HD1"
            elif atomname == "HD12":
                atomname = "HD2"
            elif atomname == "HD13":
                atomname = "HD3"
            elif atomname == "HG12":
                atomname = "HG11"
            elif atomname == "HG13":
                atomname = "HG12"
        elif resname == "CYS" and "HG" not in residue.map:
            resname = "CYS"
            if atomname == "CB":
                resname = "DISU"
                atomname = "1CB"
            elif atomname == "SG":
                resname = "DISU"
                atomname = "1SG"
        elif resname == "HIS":
            if "HD1" in residue.map and "HE2" in residue.map:
                resname = "HSP"
            elif "HD1" in residue.map:
                resname = "HSD"
            elif "HE2" in residue.map:
                resname = "HSE"
        elif resname in ["GLU", "GLH"]:
            if "HE1" in residue.map:
                if atomname == "HE1":
                    atomname = "HE2"
                elif atomname == "OE1":
                    atomname = "OE2"
                elif atomname == "OE2":
                    atomname = "OE1"
                if atomname in [
                    "CG",
                    "HG3",
                    "HG1",
                    "HG2",
                    "CD",
                    "OE1",
                    "OE2",
                    "HE2",
                ]:
                    resname = "GLUP"
                else:
                    resname = "GLU"
            elif "HE2" in residue.map:
                if atomname in [
                    "CG",
                    "HG3",
                    "HG1",
                    "HG2",
                    "CD",
                    "OE1",
                    "OE2",
                    "HE2",
                ]:
                    resname = "GLUP"
                else:
                    resname = "GLU"
        elif resname in ["ASP", "ASH"]:
            if "HD1" in residue.map:
                if atomname == "HD1":
                    atomname = "HD2"
                elif atomname == "OD1":
                    atomname = "OD2"
                elif atomname == "OD2":
                    atomname = "OD1"
                if atomname in [
                    "CB",
                    "HB3",
                    "HB1",
                    "HB2",
                    "CG",
                    "OD1",
                    "OD2",
                    "HD2",
                ]:
                    resname = "ASPP"
                else:
                    resname = "ASP"
            elif "HD2" in residue.map:
                if atomname in [
                    "CB",
                    "HB3",
                    "HB1",
                    "HB2",
                    "CG",
                    "OD1",
                    "OD2",
                    "HD2",
                ]:
                    resname = "ASPP"
                else:
                    resname = "ASP"
        # HETATM Substitutions
        if resname == "ACE":
            if atomname == "CH3":
                atomname = "CAY"
            elif atomname == "HH31":
                atomname = "HY1"
            elif atomname == "HH32":
                atomname = "HY2"
            elif atomname == "HH33":
                atomname = "HY3"
            elif atomname == "C":
                atomname = "CY"
            elif atomname == "O":
                atomname = "OY"
        elif resname == "ADP":
            atomname = atomname.replace("*", "'")
        elif resname == "NME":
            resname = "CT3"
            if atomname == "HH31":
                atomname = "HT1"
            elif atomname == "HH32":
                atomname = "HT2"
            elif atomname == "HH33":
                atomname = "HT3"
            elif atomname == "CH3":
                atomname = "CAT"
            elif atomname == "N":
                atomname = "NT"
            elif atomname == "H":
                atomname = "HNT"
        # Hydrogen Substitutions
        if atomname == "H":
            atomname = "HN"
        elif atomname == "HA2":
            atomname = "HA1"
        elif atomname == "HA3":
            atomname = "HA2"
        elif atomname == "HB2" and resname not in ["ALA"]:
            atomname = "HB1"
        elif atomname == "HB3" and resname not in ["ALA"]:
            atomname = "HB2"
        elif atomname == "HD2" and resname not in [
            "HSP",
            "HSE",
            "HSD",
            "ASPP",
        ]:
            atomname = "HD1"
        elif atomname == "HD3" and resname not in ["HIS", "HSE", "HSD"]:
            atomname = "HD2"
        elif atomname == "HE2" and resname not in [
            "TRP",
            "HSP",
            "HSE",
            "HSD",
            "GLUP",
        ]:
            atomname = "HE1"
        elif atomname == "HE3" and resname not in ["TRP", "HSP", "HSE", "HSD"]:
            atomname = "HE2"
        elif atomname == "HG2":
            atomname = "HG1"
        elif atomname == "HG3":
            atomname = "HG2"
        elif atomname == "HG" and resname in ["SER", "CYS"]:
            atomname = "HG1"
        return resname, atomname


class ForcefieldResidue:
    """ForcefieldResidue class

    The ForceFieldResidue class contains a mapping of all atoms within the
    residue for easy searching.
    """

    def __init__(self, name):
        """Initialize the ForceFieldResidue object.

        :param name:  the name of the residue
        :type name:  str
        """
        self.name = name
        self.atoms = {}

    def add_atom(self, atom):
        """Add an atom to the ForcefieldResidue object.

        :param atom:  the atom to be added
        :type atom:  Atom
        """
        atomname = atom.name
        self.atoms[atomname] = atom

    def has_atom(self, atomname):
        """Check to see if the named atom is in the current residue.

        :param atomname:  the name of the atom to search for
        :type atomname:  str
        :return:  indication of whether atmo is present
        :rtype:  bool
        """
        return atomname in self.atoms

    def get_atom(self, atomname):
        """Return the atom object with the given atomname.

        :param resname:  the name of the atom
        :type resname:  str
        :return: the atom object
        :rtype:  ForcefieldAtom
        """
        if self.has_atom(atomname):
            return self.atoms[atomname]
        return None


class ForcefieldAtom:
    """ForcefieldAtom class.

    Contains fields that are related to the forcefield at the atom level.
    """

    def __init__(self, name, charge, radius, resname, group=""):
        """Initialize the object.

        :param name:  atom name
        :type name:  str
        :param charge:  the charge on the atom
        :type charge:  float
        :param radius:  the radius of the atom
        :type radius:  float
        :param resname:  the residue name
        :type resname:  str
        :param group:  the group name
        :type group:  str
        """
        self.name = name
        self.charge = charge
        self.radius = radius
        self.resname = resname
        self.group = group

    def get(self, name):
        """Get a member of the ForcefieldAtom class.

        :param name:  the name of the member, including:

            * name:    The atom name (returns string)
            * charge:  The charge on the atom (returns float)
            * radius:  The radius of the atom (returns float)
            * epsilon: The epsilon assocaited with the atom (returns float)

        :type name:  str
        :return:  the value of the member
        :raises KeyError:  if member does not exist
        """
        try:
            return getattr(self, name)
        except AttributeError as e:
            message = (
                f'Unable to access object "{name}" in class ForcefieldAtom'
            )
            raise KeyError(message) from e

    def __str__(self):
        txt = f"{self.name}:\n"
        txt += f"  Charge: {self.charge:.4f}\n"
        txt += f"  Radius: {self.radius:.4f}"
        return txt



"""Implements the PEOE method.

The PEOE method is described in:  Paul Czodrowski, Ingo Dramburg,
Christoph A. Sotriffer  Gerhard Klebe. Development, validation, and
application of adapted PEOE charges to estimate pKa values of functional
groups in protein-ligand complexes. Proteins, 65, 424-437, 2006.
https://doi.org/10.1002/prot.21110
"""

import logging
from math import isclose

_LOGGER = logging.getLogger(__name__)


# The terms of the third-order polynomial fit for the electronegativity.
# See https://doi.org/10.1002/prot.21110 for more information.
# NOTE - this data has no meaning outside of this module; do not move.
POLY_TERMS = {
    "H": (7.17, 6.24, -0.56, 12.85),
    "C.3": (7.98, 9.18, 1.88, 19.04),
    "C.CAT": (7.98, 9.18, 1.88, 19.04),
    "C.2": (8.79 + 0.5, 9.32, 1.51, 19.62),
    "C.AR": (7.98 + 0.55, 9.18, 1.88, 19.04),
    "C.1": (10.39, 9.45, 0.73, 20.57),
    "N.3": (11.54 + 6.0, 10.28, 1.36, 28.00),
    "N.4": (11.54 + 6.0, 10.28, 1.36, 28.00),
    "N.AR": (12.87 - 1.29, 11.15, 0.85, 24.87),
    "N.2": (12.87, 11.15, 0.85, 24.87),
    "N.PL3": (12.87 + 0.5, 11.15, 0.85, 24.87),
    "N.AM": (12.87 + 3.5, 11.15, 0.85, 24.87),
    "N.1": (15.68, 11.70, -0.27, 27.11),
    "O.OH": (14.18 + 0.8, 12.92, 1.39, 28.49),
    "O.3": (14.18 - 3.1, 12.92, 1.39, 28.49),
    "O.2": (14.18, 12.92, 1.39, 28.49),
    "O.CO2": (15.25, 13.79, 0.47, 31.33),
    "F": (12.36, 13.85, 2.31, 30.82),
    "CL": (9.38 + 1.0, 9.69, 1.35, 22.04),
    "BR": (10.08 + 0.8, 8.47, 1.16, 19.71),
    "I": (9.90 + 1.0, 7.96, 0.96, 18.82),
    "S.3": (10.13 + 0.5, 9.13, 1.38, 20.65),
    "S.2": (10.13 + 0.5, 9.13, 1.38, 20.65),
    "S.O2": (10.13 + 0.5, 9.13, 1.38, 20.65),
    "P.3": (10.13 + 0.5, 9.13, 1.38, 20.65),
}
# Maximum (absolute) value of charge after which contribution to polynomial
# is capped
MAX_CHARGE = 1.1
DEFAULT_H_ELECTRONEG = 20.02
DEFAULT_H_CHARGE = 1.0
# These next values are from the "Adaptation of the PEOE Procedure" section of
# https://doi.org/10.1002/prot.21110.
DAMPING_FACTOR = 0.778
SCALING_FACTOR = 1.56
NUM_CYCLES = 6


def electronegativity(charge, poly_terms, atom_type):
    """Calculate the electronegativity.

    Calculation is based on a third-order polynomial in the atomic charge as
    described in Equation 2 of https://doi.org/10.1002/prot.21110.

    :param charge:  charge of atom
    :type charge:  float
    :param poly_terms:  polynomial terms ordered from 0th- to 3rd-order
    :param atom_type:  string with atom type
    :type atom_type:  str
    :return: electronegativity value
    :rtype:  float
    :raises IndexError:  if incorrect number of poly_terms given
    """
    chi = None
    if abs(charge) > MAX_CHARGE:
        charge = -1.0 * MAX_CHARGE if charge < 0 else MAX_CHARGE
    if (atom_type == "H") and isclose(charge, DEFAULT_H_CHARGE):
        chi = DEFAULT_H_ELECTRONEG
    else:
        if len(poly_terms) == 4:
            chi = (
                poly_terms[0]
                + poly_terms[1] * charge
                + poly_terms[2] * charge * charge
                + poly_terms[3] * charge * charge * charge
            )
        elif len(poly_terms) == 3:
            chi = (
                poly_terms[0]
                + poly_terms[1] * charge
                + poly_terms[2] * charge * charge
            )
        else:
            err = f"Cannot parse length-{len(poly_terms):d} polynomial"
            raise IndexError(err)
    return chi


def assign_terms(atoms, term_dict):
    """Assign polynomial terms to each atom.

    :param atoms:  list of Mol2Atom atoms
    :type atoms:  list
    :param term_dict:  dictionary of polynomial terms
    :type term_dict:  dict
    :return:  modified list of atoms
    :rtype:  list
    """
    for atom in atoms:
        atom_type = atom.type.upper()
        if atom_type == "O.3":
            atom_type = "O.OH"
        try:
            atom.poly_terms = term_dict[atom_type]
        except KeyError as e:
            raise KeyError(
                f"Unable to find polynomial terms for atom type {atom_type}"
            ) from e
    return atoms


def equilibrate(
    atoms,
    damp=DAMPING_FACTOR,
    scale=SCALING_FACTOR,
    num_cycles=NUM_CYCLES,
    term_dict=POLY_TERMS,
):
    """Equilibrate the atomic charges.

    :param atoms:  list of Mol2Atom atoms to equilibrate
    :type atoms:  list
    :param damp:  damping factor for equilibration process
    :type damp:  float
    :param scale:  scaling factor for equilibration process
    :type scale:  float
    :param num_cycles:  number of PEOE cycles
    :type num_cycles:  int
    :param term_dict:  dictionary of polynomial terms
    :type term_dict:  dict
    :return: revised list of atoms
    :rtype:  list
    """
    atoms = assign_terms(atoms, term_dict)
    # Reset or accumulate charges
    abs_qges = 0.0
    for atom in atoms:
        if isclose(atom.charge, 0.0):
            atom.equil_formal_charge = 0.0
        else:
            # PEOE multiples all atoms by a scaling factor at the end to
            # account for increased polarizability.  The initial formal
            # charge needs to be reduced to account for this scaling.
            atom.equil_formal_charge = atom.charge * (1.0 / scale)
            abs_qges += abs(atom.charge)
        atom.charge = 0
    # A finite number of cycles is used to prevent complete equilibration of
    # the molecule.  I'm not sure why this is a good idea but people have
    # been doing it since the original 1978 Tetrahedron paper with Gasteiger
    # & Marsili
    for icycle in range(num_cycles):
        for atom1 in atoms:
            chi1 = electronegativity(
                atom1.charge, atom1.poly_terms, atom1.type
            )
            atom1.delta_charge = 0.0
            for atom2 in atom1.bonded_atoms:
                chi2 = electronegativity(
                    atom2.charge, atom2.poly_terms, atom2.type
                )
                chi_diff = chi2 - chi1
                if chi2 > chi1:
                    chi_norm = electronegativity(
                        +1, atom1.poly_terms, atom1.type
                    )
                else:
                    chi_norm = electronegativity(
                        +1, atom2.poly_terms, atom2.type
                    )
                # Damping is used in PEOE to accelerate convergence
                atom1.delta_charge += (chi_diff / chi_norm) * (
                    damp ** (icycle + 1)
                )
        for atom in atoms:
            if isclose(abs_qges, 0.0):
                atom.charge += atom.delta_charge
            else:
                atom.charge += (
                    atom.delta_charge
                    + (1.0 / num_cycles) * atom.equil_formal_charge
                )
    for atom in atoms:
        atom.charge = scale * atom.charge
    return atoms



"""Test log caching file in the same directory as the pqr output location"""

import common
import pytest
from testfixtures import log_capture


@log_capture()
@pytest.mark.parametrize(
    "input_file,output_file",
    [
        ("1FAS.pdb", "1FAS_pdb.pqr"),
        ("1A1P.pdb", "1A1P_assign-only_whitespace_ff=AMBER_log.pqr"),
    ],
    ids=str,
)
def test_log_output_in_pqr_location(
    capture, input_file, output_file, tmp_path
):
    """Test that a log file is generated when calling the __init__.py"""
    args = "--log-level=INFO --ff=AMBER"
    input_path = common.DATA_DIR / input_file
    output_pqr = output_file
    common.run_pdb2pqr_for_tests(
        args=args,
        input_pdb=input_path,
        output_pqr=output_pqr,
        tmp_path=tmp_path,
    )

    # This match tests that the log file is output to the same location
    # as the output pqr
    capture.check_present(
        (
            "pdb2pqr.io",
            "INFO",
            f"Logs stored: {tmp_path / output_file.split('.')[0]}.log",
        )
    )



"""CIF parsing methods.

This methods use the pdbx/cif parser provided by WWPDB
(http://mmcif.wwpdb.org/docs/sw-examples/python/html/index.html)

.. todo:  Why do we have this module when we have pdbx?
.. codeauthor::  Juan Brandi
"""

import logging
from datetime import datetime

import pdbx
from numpy import ceil, minimum

from . import pdb

_LOGGER = logging.getLogger(__name__)


def atom_site(block: pdbx.containers.ContainerBase):
    """Handle ATOM_SITE block.

    Data items in the ATOM_SITE category record details about the atom sites
    in a macromolecular crystal structure, such as the positional coordinates,
    atomic displacement parameters, magnetic moments and directions.
    (Source: https://j.mp/2Zprx41)

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.ATOM objects, array of things that weren't handled
        by parser)
    :rtype:  ([Atom], [str])
    """
    line = 0
    pdb_arr = []
    err_arr = []
    atoms = block.get_object("atom_site")
    if not isinstance(atoms, pdbx.containers.DataCategory):
        _LOGGER.error(f"atom_site: No lines were found\n")
        raise ValueError(
            "Did not find atom_site in cif file, terminal failure."
        )
    num_model_arr = count_models(block)
    if len(num_model_arr) == 1:
        # TODO - this part of the conditional should be a separate function
        for i in range(atoms.row_count):
            line = convert_cif_atom_site_to_pdb_line(atoms=atoms, row_index=i)
            if line is None:
                continue
            try:
                if atoms.get_value("group_PDB", i) == "ATOM":
                    pdb_arr.append(pdb.ATOM(line))
                elif atoms.get_value("group_PDB", i) == "HETATM":
                    pdb_arr.append(pdb.HETATM(line))
            except KeyError:
                _LOGGER.error(f"atom_site: Error reading line: #{line}#\n")
        return pdb_arr, err_arr
    # TODO - Given the return statement above, is this "else" ever reached?
    else:
        # TODO - this part of the conditional should be a separate function
        for j in num_model_arr:
            try:
                line = "MODEL "
                line += " " * 4
                line += " " * (4 - len(str(j))) + str(j)
                pdb_arr.append(pdb.MODEL(line))
            except KeyError:
                _LOGGER.error(f"atom_site: Error readline line:\n{line}")
                err_arr.append("MODEL")

            for i in range(atoms.row_count):
                if atoms.get_value("pdbx_PDB_model_num", i) == j:
                    line = convert_cif_atom_site_to_pdb_line(
                        atoms=atoms, row_index=i
                    )
                    if line is None:
                        continue
                    try:
                        if atoms.get_value("group_PDB", i) == "ATOM":
                            pdb_arr.append(pdb.ATOM(line))
                        elif atoms.get_value("group_PDB", i) == "HETATM":
                            pdb_arr.append(pdb.HETATM(line))
                    except (KeyError, ValueError):
                        _LOGGER.error(
                            f"atom_site: Error reading line: #{line}#\n"
                        )
            try:
                line = "ENDMDL"
                pdb_arr.append(pdb.ENDMDL(line))
            except KeyError:
                _LOGGER.error(f"atom_site: Error reading line:\n{line}")
                err_arr.append("ENDMDL")
        return pdb_arr, err_arr


def convert_cif_atom_site_to_pdb_line(
    atoms: pdbx.containers.DataCategory,
    row_index: int,
) -> str | None:
    """Converts cif data for one row into a pdb line.

    Extracts all the relevant fields from atoms, does basic formatting,
    and converts them into a pdb line.

    :param atoms: DataCategory containing all atoms.
    :type atoms: pdbx.containers.DataCategory
    :param row_index: Index corresponding to the cif line we want to convert to a pdb line
    :type row_index: int
    :return: Pdb line (exactly 80 chars long), will be None if it failed to process it properly
    :rtype: str | None
    """
    # Extract and cast values
    group = atoms.get_value("group_PDB", row_index)
    serial = int(atoms.get_value("id", row_index))
    name = atoms.get_value("label_atom_id", row_index)
    alt_id = atoms.get_value("label_alt_id", row_index)
    res_name = atoms.get_value("label_comp_id", row_index)
    chain = atoms.get_value("label_asym_id", row_index)
    res_seq = int(atoms.get_value("auth_seq_id", row_index))
    x = float(atoms.get_value("Cartn_x", row_index))
    y = float(atoms.get_value("Cartn_y", row_index))
    z = float(atoms.get_value("Cartn_z", row_index))
    occ_raw = atoms.get_value("occupancy", row_index)
    try:
        occ = float(occ_raw)
    except (TypeError, ValueError):
        _LOGGER.error(
            f"Invalid occupancy {occ_raw} at row {row_index}, setting to 1.0"
        )
        occ = 1.0
    temp_raw = atoms.get_value("B_iso_or_equiv", row_index)
    try:
        temp = float(temp_raw)
    except (TypeError, ValueError):
        _LOGGER.error(
            f"Invalid B-factor {temp_raw} at row {row_index}, setting to 0.0"
        )
        temp = 0.0
    element = atoms.get_value("type_symbol", row_index)

    # Handle the '?' or '.' cases for alt_id and charge
    alt_id = (
        alt_id if (alt_id is not None and alt_id not in (".", "?")) else " "
    )
    charge = (
        atoms.get_value("pdbx_formal_charge", row_index)
        if "pdbx_formal_charge" in atoms.attribute_list
        else "  "
    )
    if charge in ["?", None, "."]:
        charge = "  "

    # Format the Atom Name (4 chars, specifically aligned)
    # 1-2 char elements start at the 2nd position of the 4-char field
    if len(element) < 2 and len(name) < 4:
        name = f" {name:<3}"
    else:
        name = f"{name:<4}"

    # THE PDB LINE MAPPING
    line = (
        f"{group:<6}"  #    1-6 Record name
        f"{serial:>5}"  #   7-11 Atom serial number
        f" "  #             12 Space
        f"{name}"  #        13-16 Atom name
        f"{alt_id:1}"  #    17 Alternate location indicator
        f"{res_name:>3}"  # 18-20 Residue name
        f" "  #             21 Space
        f"{chain:1}"  #     22 Chain Id
        f"{res_seq:>4}"  #  23-26 Residue sequence number
        f" "  #             27 Code for insertion of residues
        f"   "  #           28-30 Spaces
        f"{x:8.3f}"  #      31-38 X-coordinates
        f"{y:8.3f}"  #      39-46 Y-coordinates
        f"{z:8.3f}"  #      47-54 Z-coordinates
        f"{occ:6.2f}"  #    55-60 Occupancy
        f"{temp:6.2f}"  #   61-66 Temperature factor
        f"          "  #    67-76 Spaces
        f"{element:>2}"  #  77-78 Element Symbol
        f"{charge:>2}"  #   79-80 Charge of the atom
    )
    if len(line) != 80:
        # This would happen if any of the values above are larger than the max space they are allocated, e.g.:
        # if x < -999.999
        # if name = LIG1
        _LOGGER.error(
            f"atom_site: pdb line extracted from cif record is not exactly 80 char long, dropping line:\n{line}"
        )
        return None
    return line


def conect(block):
    """Handle CONECT block.

    Data items in the STRUCT_CONN category record details about the connections
    between portions of the structure.
    These can be hydrogen bonds, salt bridges, disulfide bridges and so on.

    The ``STRUCT_CONN_TYPE`` records define the criteria used to identify these
    connections.
    (Source: https://j.mp/3gPkJT5)

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    pdb_arr = []
    err_arr = []
    struct_conn = block.get_object("struct_conn")
    atoms = block.get_object("atom_site")
    if struct_conn is None or atoms is None:
        return pdb_arr, err_arr
    for index in range(struct_conn.row_count):
        atom_pair = []
        for partner in ["ptnr1_", "ptnr2_"]:
            # Retrieve all the information necessary to uniquely identify atom4
            atom_dict = {
                "auth_seq_id": struct_conn.get_value(
                    partner + "auth_seq_id", index
                ),
                "auth_comp_id": struct_conn.get_value(
                    partner + "auth_comp_id", index
                ),
                "auth_asym_id": struct_conn.get_value(
                    partner + "auth_asym_id", index
                ),
                "label_atom_id": struct_conn.get_value(
                    partner + "label_atom_id", index
                ),
            }
            for i in range(atoms.row_count):
                found = all(
                    atoms.get_value(key, i) == atom_dict[key]
                    for key in atom_dict
                )
                if found:
                    atom_pair.append(atoms.get_value("id", i))
        if len(atom_pair) == 2:
            pline = (
                "CONECT"
                + " " * (5 - len(str(atom_pair[0])))
                + str(atom_pair[0])
                + " " * (5 - len(str(atom_pair[1])))
                + str(atom_pair[1])
            )
            try:
                pdb_arr.append(pdb.CONECT(pline))
            except KeyError:
                _LOGGER.error(f"conect:   Error parsing line: \n{pline}")
                err_arr.append("conect")
    return pdb_arr, err_arr


def header(block):
    """Handle HEADER block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    header_arr = []
    header_err = []

    struct_obj = block.get_object("struct_keywords")
    database_obj = block.get_object("pdbx_database_status")
    entry_obj = block.get_object("entry")
    line = "HEADER"
    line += " " * 4
    if struct_obj is not None:
        line += struct_obj.get_value("pdbx_keywords") + " " * (
            40 - len(struct_obj.get_value("pdbx_keywords"))
        )
    else:
        line += " " * 40
    if database_obj is not None:
        ridd = database_obj.get_value("recvd_initial_deposition_date")
        if isinstance(ridd, str) and len(ridd) > 9:
            ridd = (
                datetime.strptime(ridd, "%Y-%m-%d")
                .strftime("%d-%b-%y")
                .upper()
            )
            line += " " * (9 - len(ridd)) + ridd
    else:
        line += " " * 9
    line += " " * 3
    if entry_obj is not None:
        line += " " * (
            4 - len(entry_obj.get_value("id"))
        ) + entry_obj.get_value("id")
    else:
        line += " " * 4
    try:
        header_arr.append(pdb.HEADER(line))
    except KeyError:
        _LOGGER.error(f"header:   Error parsing line: #{line}#")
        header_err.append("header")
    return header_arr, header_err


def title(block):
    """Handle TITLE block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    title_arr = []
    title_err = []
    struct_obj = block.get_object("struct")
    title_string = struct_obj.get_value("title")
    title_chunk = int(ceil(len(title_string) / 70.0))
    for i in range(title_chunk):
        line = "TITLE  "
        line += " " * (2 - len(str(i + 1))) + str(i + 1) if i > 0 else "  "
        line += title_string[
            (i * 70) : minimum(len(title_string), (i + 1) * 70)
        ]
        try:
            title_arr.append(pdb.TITLE(line))
        except KeyError:
            _LOGGER.error(f"TITLE:    Error parsing line:\n{line}")
            title_err.append("title")
    return title_arr, title_err


def compnd(block):
    """Handle COMPND block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    compnd_arr = []
    compnd_err = []
    entity_obj = block.get_object("entity")
    cont = 1
    for i in range(entity_obj.row_count):
        line1 = "COMPND "
        line1 += " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
        line1 += "MOL_ID: " + str(entity_obj.get_value("id", i)) + ""
        try:
            compnd_arr.append(pdb.COMPND(line1))
        except KeyError:
            _LOGGER.error(f"compnd:    Error parsing line:\n{line1}")
            compnd_err.append("compnd")
        cont += 1
        line2 = "COMPND "
        line2 += " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
        line2 += (
            "MOLECULE: " + entity_obj.get_value("pdbx_description", i) + ""
        )
        try:
            compnd_arr.append(pdb.COMPND(line2))
        except KeyError:
            _LOGGER.error(f"compnd:    Error parsing line:\n{line2}")
            compnd_err.append("compnd")
        cont += 1
    return compnd_arr, compnd_err


def source(block: pdbx.containers.ContainerBase):
    """Handle SOURCE block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    src_arr = []
    src_err = []
    src_obj = block.get_object("entity_src_gen")
    if src_obj is None:
        return src_arr, src_err
    cont = 1
    for i in range(src_obj.row_count):
        if ("entity_id" in src_obj.attribute_list) and (
            src_obj.get_value("entity_id", i) not in ["?", None]
        ):
            line = "SOURCE "
            line += (
                " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
            )
            line += "MOL_ID: " + str(src_obj.get_value("entity_id", i)) + ""
            cont += 1
            try:
                src_arr.append(pdb.SOURCE(line))
            except KeyError:
                _LOGGER.error(f"source:  Error parsing line:\n{line}")
                src_err.append("source")
        if ("pdbx_gene_src_scientific_name" in src_obj.attribute_list) and (
            src_obj.get_value("pdbx_gene_src_scientific_name", i)
            not in ["?", None]
        ):
            line = "SOURCE "
            line += (
                " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
            )
            line += (
                "ORGANISM_SCIENTIFIC: "
                + src_obj.get_value("pdbx_gene_src_scientific_name", i)
                + ""
            )
            cont += 1
            try:
                src_arr.append(pdb.SOURCE(line))
            except KeyError:
                _LOGGER.error(f"source:  Error parsing line:\n{line}")
                src_err.append("source")
        if ("gene_src_common_name" in src_obj.attribute_list) and (
            src_obj.get_value("gene_src_common_name", i) not in ["?", None]
        ):
            line = "SOURCE "
            line += (
                " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
            )
            line += (
                "ORGANISM_COMMON: "
                + src_obj.get_value("gene_src_common_name", i)
                + ""
            )
            cont += 1
            try:
                src_arr.append(pdb.SOURCE(line))
            except KeyError:
                _LOGGER.error(f"source:  Error parsing line:\n{line}")
                src_err.append("source")
        if ("pdbx_gene_src_ncbi_taxonomy_id" in src_obj.attribute_list) and (
            src_obj.get_value("pdbx_gene_src_ncbi_taxonomy_id", i)
            not in ["?", None]
        ):
            line = "SOURCE "
            line += (
                " " * (3 - len(str(cont))) + str(cont) if cont > 1 else "   "
            )
            line += (
                "ORGANISM_TAXID: "
                + src_obj.get_value("pdbx_gene_src_ncbi_taxonomy_id", i)
                + ""
            )
            cont += 1
            try:
                src_arr.append(pdb.SOURCE(line))
            except KeyError:
                _LOGGER.error(f"source:    Error parsing line:\n{line}")
                src_err.append("source")
    return src_arr, src_err


def keywds(block):
    """Handle KEYWDS block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    key_arr = []
    key_err = []
    key_obj = block.get_object("struct_keywords")
    if key_obj is None:
        return key_arr, key_err
    key_string = key_obj.get_value("text")
    key_chunk = int(ceil(len(key_string) / 69.0))
    for i in range(key_chunk):
        line = "KEYWDS  "
        line += " " * (2 - len(str(i + 1))) + str(i + 1) if i > 0 else "  "
        line += key_string[(i * 69) : minimum(len(key_string), (i + 1) * 69)]
        try:
            key_arr.append(pdb.KEYWDS(line))
        except KeyError:
            _LOGGER.error(f"keywds:    Error parsing line:\n{line}")
            key_err.append("keywds")
    return key_arr, key_err


def expdata(block):
    """Handle EXPDTA block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    ex_arr = []
    ex_err = []
    ex_obj = block.get_object("exptl")
    if ex_obj is None:
        return ex_arr, ex_err
    line = "EXPDTA  "
    line += "  "
    line += ex_obj.get_value("method", 0)
    try:
        ex_arr.append(pdb.EXPDTA(line))
    except KeyError:
        _LOGGER.error(f"expdata:    Error parsing line:\n{line}\n")
        ex_err.append("expdata")
    return ex_arr, ex_err


def author(block):
    """Handle AUTHOR block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    aut_arr = []
    aut_err = []
    aut_obj = block.get_object("audit_author")
    if aut_obj is None:
        return aut_arr, aut_err
    for i in range(aut_obj.row_count):
        line = "AUTHOR  "
        line += "  " * (
            2 - len(str(aut_obj.get_value("pdbx_ordinal", i)))
        ) + str(aut_obj.get_value("pdbx_ordinal", i))
        line += aut_obj.get_value("name", i)
        try:
            aut_arr.append(pdb.AUTHOR(line))
        except KeyError:
            _LOGGER.error(f"author:  Error parsing line:\n{line}")
            aut_err.append("author")
    return aut_arr, aut_err


def cryst1(block):
    """Handle CRYST1 block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    cry_arr = []
    cry_err = []
    cry_obj = block.get_object("cell")
    sym_obj = block.get_object("symmetry")
    if cry_obj is None:
        return cry_arr, cry_err
    line = "CRYST1"
    line += " " * (
        9 - len(str(cry_obj.get_value("length_a", 0)))
    ) + cry_obj.get_value("length_a", 0)
    line += " " * (
        9 - len(str(cry_obj.get_value("length_b", 0)))
    ) + cry_obj.get_value("length_b", 0)
    line += " " * (
        9 - len(str(cry_obj.get_value("length_c", 0)))
    ) + cry_obj.get_value("length_c", 0)
    line += " " * (
        7 - len(str(cry_obj.get_value("angle_alpha", 0)))
    ) + cry_obj.get_value("angle_alpha", 0)
    line += " " * (
        7 - len(str(cry_obj.get_value("angle_beta", 0)))
    ) + cry_obj.get_value("angle_beta", 0)
    line += " " * (
        7 - len(str(cry_obj.get_value("angle_gamma", 0)))
    ) + cry_obj.get_value("angle_gamma", 0)
    if sym_obj is not None:
        line += " " * (
            11 - len(str(sym_obj.get_value("space_group_name_H-M", 0)))
        ) + sym_obj.get_value("space_group_name_H-M", 0)
    else:
        line += " " * 11
    line += " " * (
        4 - len(str(cry_obj.get_value("Z_PDB", 0)))
    ) + cry_obj.get_value("Z_PDB", 0)
    try:
        cry_arr.append(pdb.CRYST1(line))
    except KeyError:
        _LOGGER.error(f"cif.cryst1:    Error parsing line:\n{line}")
        cry_err.append(cryst1)
    return cry_arr, cry_err


def scalen(block):
    """Handle SCALEn block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    sc_arr = []
    sc_err = []
    sc_obj = block.get_object("atom_sites")
    if sc_obj is None:
        return sc_arr, sc_err
    scale1 = ""
    scale1 += "SCALE1    "
    scale1 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[1][1]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[1][1]", 0))
    scale1 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[1][2]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[1][2]", 0))
    scale1 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[1][3]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[1][3]", 0))
    scale1 += "     "
    scale1 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_vector[1]", 0)))
    ) + str(sc_obj.get_value("fract_transf_vector[1]", 0))
    scale2 = ""
    scale2 += "SCALE2    "
    scale2 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[2][1]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[2][1]", 0))
    scale2 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[2][2]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[2][2]", 0))
    scale2 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[2][3]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[2][3]", 0))
    scale2 += "     "
    scale2 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_vector[2]", 0)))
    ) + str(sc_obj.get_value("fract_transf_vector[2]", 0))
    scale3 = ""
    scale3 += "SCALE3    "
    scale3 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[3][1]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[3][1]", 0))
    scale3 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[3][2]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[3][2]", 0))
    scale3 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_matrix[3][3]", 0)))
    ) + str(sc_obj.get_value("fract_transf_matrix[3][3]", 0))
    scale3 += "     "
    scale3 += " " * (
        10 - len(str(sc_obj.get_value("fract_transf_vector[3]", 0)))
    ) + str(sc_obj.get_value("fract_transf_vector[3]", 0))
    try:
        sc_arr.append(pdb.SCALE1(scale1))
    except KeyError:
        _LOGGER.error(f"cif.scalen:    Error parsing line:\n{scale1}")
        sc_err.append("SCALE1")
    try:
        sc_arr.append(pdb.SCALE2(scale2))
    except KeyError:
        _LOGGER.error(f"cif.scalen:    Error parsing line:\n{scale2}")
        sc_err.append("SCALE2")
    try:
        sc_arr.append(pdb.SCALE3(scale3))
    except KeyError:
        _LOGGER.error(f"cif.scalen:    Error parsing line:\n{scale3}")
        sc_err.append("SCALE3")
    return sc_arr, sc_err


def origxn(block):
    """Handle ORIGXn block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    or_arr = []
    or_err = []
    or_obj = block.get_object("database_PDB_matrix")
    if or_obj is None:
        return or_arr, or_err
    orig1 = "ORIGX1    "
    orig1 += " " * (10 - len(str(or_obj.get_value("origx[1][1]", 0)))) + str(
        or_obj.get_value("origx[1][1]", 0)
    )
    orig1 += " " * (10 - len(str(or_obj.get_value("origx[1][2]", 0)))) + str(
        or_obj.get_value("origx[1][2]", 0)
    )
    orig1 += " " * (10 - len(str(or_obj.get_value("origx[1][3]", 0)))) + str(
        or_obj.get_value("origx[1][3]", 0)
    )
    orig1 += "     "
    orig1 += " " * (
        10 - len(str(or_obj.get_value("origx_vector[1]", 0)))
    ) + str(or_obj.get_value("origx_vector[1]", 0))
    orig2 = "ORIGX2    "
    orig2 += " " * (10 - len(str(or_obj.get_value("origx[2][1]", 0)))) + str(
        or_obj.get_value("origx[2][1]", 0)
    )
    orig2 += " " * (10 - len(str(or_obj.get_value("origx[2][2]", 0)))) + str(
        or_obj.get_value("origx[2][2]", 0)
    )
    orig2 += " " * (10 - len(str(or_obj.get_value("origx[2][3]", 0)))) + str(
        or_obj.get_value("origx[2][3]", 0)
    )
    orig2 += "     "
    orig2 += " " * (
        10 - len(str(or_obj.get_value("origx_vector[2]", 0)))
    ) + str(or_obj.get_value("origx_vector[2]", 0))
    orig3 = "ORIGX3    "
    orig3 += " " * (10 - len(str(or_obj.get_value("origx[3][1]", 0)))) + str(
        or_obj.get_value("origx[3][1]", 0)
    )
    orig3 += " " * (10 - len(str(or_obj.get_value("origx[3][2]", 0)))) + str(
        or_obj.get_value("origx[3][2]", 0)
    )
    orig3 += " " * (10 - len(str(or_obj.get_value("origx[3][3]", 0)))) + str(
        or_obj.get_value("origx[3][3]", 0)
    )
    orig3 += "     "
    orig3 += " " * (
        10 - len(str(or_obj.get_value("origx_vector[3]", 0)))
    ) + str(or_obj.get_value("origx_vector[3]", 0))
    try:
        or_arr.append(pdb.ORIGX1(orig1))
    except KeyError:
        _LOGGER.error(f"cif.origxn:  Error parsing line:\n{orig1}")
        or_err.append("ORIGX1")
    try:
        or_arr.append(pdb.ORIGX2(orig2))
    except KeyError:
        _LOGGER.error(f"cif.origxn:  Error parsing line:\n{orig2}")
        or_err.append("ORIGX2")
    try:
        or_arr.append(pdb.ORIGX3(orig3))
    except KeyError:
        _LOGGER.error(f"cif.origxn:  Error parsing line:\n{orig3}")
        or_err.append("ORIGX3")
    return or_arr, or_err


def cispep(block):
    """Handle CISPEP block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    cis_arr = []
    cis_err = []
    cis_obj = block.get_object("struct_mon_prot_cis")
    if cis_obj is None:
        return cis_arr, cis_err
    for i in range(cis_obj.row_count):
        line = "CISPEP "
        line += " " * (3 - len(str(cis_obj.get_value("pdbx_id", i)))) + str(
            cis_obj.get_value("pdbx_id", i)
        )
        line += " "
        line += " " * (
            3 - len(cis_obj.get_value("auth_comp_id", i))
        ) + cis_obj.get_value("auth_comp_id", i)
        line += " "
        line += cis_obj.get_value("auth_asym_id", i)
        line += " "
        line += " " * (
            4 - len(str(cis_obj.get_value("auth_seq_id", i)))
        ) + str(cis_obj.get_value("auth_seq_id", i))
        value = cis_obj.get_value("pdbx_PDB_ins_code", i)
        if value not in ["?", None]:
            line += value
        else:
            line += " "
        line += "   "
        line += " " * (
            3 - len(cis_obj.get_value("pdbx_auth_comp_id_2", i))
        ) + cis_obj.get_value("pdbx_auth_comp_id_2", i)
        line += " "
        line += cis_obj.get_value("pdbx_auth_asym_id_2", i)
        line += " "
        line += " " * (
            4 - len(str(cis_obj.get_value("pdbx_auth_seq_id_2", i)))
        ) + str(cis_obj.get_value("pdbx_auth_seq_id_2", i))
        value = cis_obj.get_value("pdbx_PDB_ins_code_2", i)
        if value not in ["?", None]:
            line += value
        else:
            line += " "
        line += " " * 7
        line += " " * (
            3 - len(str(cis_obj.get_value("pdbx_PDB_model_num", i)))
        ) + str(cis_obj.get_value("pdbx_PDB_model_num", i))
        line += " " * 7
        line += " " * (
            6 - len(str(cis_obj.get_value("pdbx_omega_angle", i)))
        ) + str(cis_obj.get_value("pdbx_omega_angle", i))
        try:
            cis_arr.append(pdb.CISPEP(line))
        except KeyError:
            _LOGGER.error(f"cif.cispep:    Erro parsing line:\n{line}")
            cis_err.append("cispep")
    return cis_arr, cis_err


def ssbond(block):
    """Handle SSBOND block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  (array of pdb.conect objects, array of things that did not parse)
    :rtype:  ([pdb.CONECT], [str])
    """
    ssb_arr = []
    ssb_err = []
    ssb_obj = block.get_object("struct_conn")
    if ssb_obj is None:
        return ssb_arr, ssb_err
    for i in range(ssb_obj.row_count):
        line = "SSBOND "
        line += " " * (3 - len(str(ssb_obj.get_value("id", i)[-1]))) + str(
            ssb_obj.get_value("id", i)[-1]
        )
        line += " "
        line += " " * (
            3 - len(ssb_obj.get_value("ptnr1_auth_comp_id", i))
        ) + ssb_obj.get_value("ptnr1_auth_comp_id", i)
        line += " "
        line += ssb_obj.get_value("ptnr1_auth_asym_id", i)
        line += " "
        line += " " * (
            4 - len(str(ssb_obj.get_value("ptnr1_auth_seq_id", i)))
        ) + str(ssb_obj.get_value("ptnr1_auth_seq_id", i))
        value = ssb_obj.get_value("pdbx_ptnr1_PDB_ins_code", i)
        if value not in ["?", None]:
            line += value
        else:
            line += " "
        line += " " * 3
        line += " " * (
            3 - len(ssb_obj.get_value("ptnr2_auth_comp_id", i))
        ) + ssb_obj.get_value("ptnr2_auth_comp_id", i)
        line += " "
        line += ssb_obj.get_value("ptnr2_auth_asym_id", i)
        line += " "
        line += " " * (
            4 - len(str(ssb_obj.get_value("ptnr2_auth_seq_id", i)))
        ) + str(ssb_obj.get_value("ptnr2_auth_seq_id", i))
        value = ssb_obj.get_value("pdbx_ptnr2_PDB_ins_code", i)
        if value not in ["?", None]:
            line += value
        else:
            line += " "
        line += " " * 23
        line += " " * (
            6 - len(ssb_obj.get_value("ptnr1_symmetry", i).replace("_", ""))
        ) + ssb_obj.get_value("ptnr1_symmetry", i).replace("_", "")
        line += " "
        line += " " * (
            6 - len(ssb_obj.get_value("ptnr2_symmetry", i).replace("_", ""))
        ) + ssb_obj.get_value("ptnr2_symmetry", i).replace("_", "")
        line += " "
        line += " " * (
            5 - len(str(ssb_obj.get_value("pdbx_dist_value", i)))
        ) + str(ssb_obj.get_value("pdbx_dist_value", i))
        try:
            ssb_arr.append(pdb.SSBOND(line))
        except KeyError:
            _LOGGER.error(f"cif.ssbond:    Error parsing line:\n{line}")
            ssb_err.append("ssbond")
    return ssb_arr, ssb_err


def count_models(block):
    """Count models in structure file block.

    :param block:  PDBx data block
    :type block:  [str]
    :return:  number of models in block
    :rtype:  int
    """
    atom_obj = block.get_object("atom_site")
    model_num = []
    for i in range(atom_obj.row_count):
        tmp = atom_obj.get_value("pdbx_PDB_model_num", i)
        if tmp not in model_num:
            model_num.append(tmp)
    return model_num


def read_cif(cif_file):
    """Parse CIF-format data into array of Atom objects.

    .. todo::  Manage several blocks of data.

    :param file:  open file-like object
    :type file:  file
    :return:  (a dictionary indexed by PDBx/CIF record names, a list of record
        names that couldn't be parsed)
    :rtype:  (dict, [str])
    """
    pdblist = []  # Array of parsed lines (as objects)
    errlist = []  # List of record names that couldn't be parsed.
    if cif_file is None:
        return pdblist, errlist
    pdbdata = pdbx.load(cif_file)
    if len(pdbdata) > 0:
        for block in pdbdata:
            head_pdb, head_err = header(block)
            title_pdb, title_err = title(block)
            cmpnd_pdb, cmpnd_err = compnd(block)
            src_pdb, src_err = source(block)
            key_pdb, key_err = keywds(block)
            ex_pdb, ex_err = expdata(block)
            aut_pdb, aut_err = author(block)
            ssb_pdb, ssb_err = ssbond(block)
            cis_pdb, cis_err = cispep(block)
            cry_pdb, cry_err = cryst1(block)
            or_pdb, or_err = origxn(block)
            sc_pdb, sc_err = scalen(block)
            ato_pdb, ato_err = atom_site(block)
            con_pdb, con_err = conect(block)
            pdblist = (
                head_pdb
                + title_pdb
                + cmpnd_pdb
                + src_pdb
                + key_pdb
                + ex_pdb
                + aut_pdb
                + ssb_pdb
                + cis_pdb
                + cry_pdb
                + or_pdb
                + sc_pdb
                + ato_pdb
                + con_pdb
            )
            errlist = (
                head_err
                + title_err
                + cmpnd_err
                + src_err
                + key_err
                + ex_err
                + aut_err
                + ssb_err
                + cis_err
                + cry_err
                + or_err
                + sc_err
                + ato_err
                + con_err
            )
    else:
        _LOGGER.error("Unknown error while reading CIF file.")

    return pdblist, errlist



"""PDB2PQR Version number.

Store the version here so:

* we don't load dependencies by storing it in :file:`__init__.py`
* we can import it in setup.py for the same reason
* we can import it into your module
"""

__version__ = "3.7.1.post1"



"""Create an APBS input file using :mod:`psize` data.

.. codeauthor::  Todd Dolinsky
.. codeauthor::  Nathan Baker
"""

import argparse
import logging
import pickle
from pathlib import Path

from . import psize
from .config import TITLE_STR

_LOGGER = logging.getLogger(__name__)


class Elec:
    """An object for the ELEC section of an APBS input file."""

    def __init__(
        self, pqrpath, size, method, asyncflag, istrng=0, *, potdx=False
    ):
        """Initialize object.

        .. todo::  Remove hard-coded parameters.

        :param pqrpath:  path to PQR file
        :type pqrpath:  str
        :param size:  parameter sizing object
        :type size:  Psize
        :param method:  solution method (e.g., mg-para, mg-auto, etc.)
        :type method:  str
        :param asyncflag:  perform an asynchronous parallel focusing
            calculation
        :type asyncflag:  bool
        :param istrng:  ionic strength/concentration (M)
        :type istring:  float
        :param potdx:  whether to write out potential information in DX format
        :type potdx:  bool
        """
        # If this is an async or parallel calc, we want to use
        # the per-grid dime rather than the global dime.
        self.dime = size.ngrid
        gmem = (
            200.0
            * self.dime[0]
            * self.dime[1]
            * self.dime[2]
            / 1024.0
            / 1024.0
        )
        if method == "":  # method not named - use ceiling
            if gmem > size.gmemceil:
                method = "mg-para"
            else:
                method = "mg-auto"
        if method == "mg-para":
            self.dime = size.getSmallest()
        self.method = method
        self.istrng = istrng
        self.glen = size.coarse_length
        self.cglen = size.coarse_length
        self.fglen = size.fine_length
        self.pdime = size.proc_grid
        self.label = ""
        self.nlev = 4
        self.ofrac = 0.1
        self.async_ = 0
        self.asyncflag = asyncflag
        self.cgcent = "mol 1"
        self.fgcent = "mol 1"
        self.gcent = "mol 1"
        self.mol = 1
        self.lpbe = 1
        self.npbe = 0
        self.bcfl = "sdh"
        # TODO - where did these very arbitrary numbers come from?
        self.ion = [[-1, 1.815], [1, 1.875]]  # Multiple ions possible
        self.pdie = 2.0
        self.sdie = 78.54
        self.srfm = "smol"
        self.chgm = "spl2"
        self.sdens = 10.0
        self.srad = 1.4
        self.swin = 0.3
        self.temp = 298.15
        self.gamma = 0.105
        self.calcenergy = "total"
        self.calcforce = "no"
        if potdx:
            self.write = [["pot", "dx", pqrpath]]
        else:
            # Multiple write statements possible
            self.write = [["pot", "dx", "pot"]]

    def __str__(self):
        text = f"elec {self.label}\n"
        text += f"    {self.method}\n"
        text += f"    dime {int(self.dime[0])} {int(self.dime[1])} "
        text += f"{int(self.dime[2])}\n"
        if self.method == "mg-auto":
            text += f"    cglen {self.cglen[0]:.4f} {self.cglen[1]:.4f} "
            text += f"{self.cglen[2]:.4f}\n"
            text += f"    fglen {self.fglen[0]:.4f} {self.fglen[1]:.4f} "
            text += f"{self.fglen[2]:.4f}\n"
            text += f"    cgcent {self.cgcent}\n"
            text += f"    fgcent {self.fgcent}\n"
        elif self.method == "mg-manual":
            text += f"    glen {self.glen[0]:.3f} {self.glen[1]:.3f} "
            text += f"{self.glen[2]:.3f}\n"
            text += f"    gcent {self.gcent}\n"
        elif self.method == "mg-para":
            text += f"    pdime {int(self.pdime[0])} {int(self.pdime[1])} "
            text += f"{int(self.pdime[2])}\n"
            text += f"    ofrac {self.ofrac:.1f}\n"
            text += f"    cglen {self.cglen[0]:.4f} {self.cglen[1]:.4f} "
            text += f"{self.cglen[2]:.4f}\n"
            text += f"    fglen {self.fglen[0]:.4f} {self.fglen[1]:.4f} "
            text += f"{self.fglen[2]:.4f}\n"
            text += f"    cgcent {self.cgcent}\n"
            text += f"    fgcent {self.fgcent}\n"
            if self.asyncflag:
                text += f"    async {self.async_}\n"
        text += f"    mol {int(self.mol)}\n"
        text += "    lpbe\n" if self.lpbe else "    npbe\n"
        text += f"    bcfl {self.bcfl}\n"
        if self.istrng > 0:
            for ion in self.ion:
                text += f"    ion charge {ion[0]:.2f} conc {self.istrng:.3f} "
                text += f"radius {ion[1]:.4f}\n"
        text += f"    pdie {self.pdie:.4f}\n"
        text += f"    sdie {self.sdie:.4f}\n"
        text += f"    srfm {self.srfm}\n"
        text += f"    chgm {self.chgm}\n"
        text += f"    sdens {self.sdens:.2f}\n"
        text += f"    srad {self.srad:.2f}\n"
        text += f"    swin {self.swin:.2f}\n"
        text += f"    temp {self.temp:.2f}\n"
        text += f"    calcenergy {self.calcenergy}\n"
        text += f"    calcforce {self.calcforce}\n"
        for write in self.write:
            text += f"    write {write[0]} {write[1]} {write[2]}\n"
        text += "end\n"
        return text


class Input:
    """Each object of this class is one APBS input file."""

    def __init__(
        self, pqrpath, size, method, asyncflag, istrng=0, *, potdx=False
    ):
        """Initialize the input file class.

        Each input file contains a PQR name, a list of elec objects, and a
        list of strings containing print statements.
        For starters, assume two ELEC statements are needed, one for the
        inhomgenous and the other for the homogenous dielectric calculations.

        .. note::
            This assumes you have already run psize, either by
            :func:`size.run_psize(...)` or :func:`size.parse_string(...)`
            followed by :func:`size.set_all()`.

        :param pqrpath:  path to PQR file
        :type pqrpath:  str
        :param size:  parameter sizing object
        :type size:  Psize
        :param method:  solution method (e.g., mg-para, mg-auto, etc.)
        :type method:  str
        :param asyncflag:  perform an asynchronous parallel focusing
            calculation
        :type asyncflag:  bool
        :param istrng:  ionic strength/concentration (M)
        :type istring:  float
        :param potdx:  whether to write out potential information in DX format
        :type potdx:  bool
        """
        self.pqrpath = Path(pqrpath)
        self.pqrname = self.pqrpath.name
        self.asyncflag = asyncflag
        # Initialize variables to default elec values
        elec1 = Elec(pqrpath, size, method, asyncflag, istrng, potdx=potdx)
        if not potdx:
            elec2 = Elec(pqrpath, size, method, asyncflag, istrng, potdx=potdx)
            elec2.sdie = 2.0
            elec2.write = []
        else:
            elec2 = ""
        self.elecs = [elec1, elec2]
        if not potdx:
            self.prints = ["print elecEnergy 2 - 1 end"]
        else:
            self.prints = ["print elecEnergy 1 end"]

    def __str__(self):
        text = "read\n"
        text += f"    mol pqr {self.pqrname}\n"
        text += "end\n"
        for elec in self.elecs:
            text += str(elec)
        for prints in self.prints:
            text += prints
        text += "\nquit\n"
        return text

    def print_input_files(self, output_path):
        """Generate the input file(s) associated with this object.

        :param output_path:  location for generated files
        :type output_path:  str
        """
        path = Path(output_path)
        base_name = path.stem
        if self.asyncflag:
            outname = base_name + "-para.in"
            # Temporarily disable async flag
            for elec in self.elecs:
                elec.asyncflag = False
            with open(outname, "w") as out_file:
                out_file.write(str(self))
            # Now make the async files
            elec = self.elecs[0]
            nproc = elec.pdime[0] * elec.pdime[1] * elec.pdime[2]
            for i in range(int(nproc)):
                outname = base_name + f"-PE{i}.in"
                for elec in self.elecs:
                    elec.asyncflag = True
                    elec.async_ = i
                with open(outname, "w") as out_file:
                    out_file.write(str(self))
        else:
            with open(path, "w") as out_file:
                out_file.write(str(self))

    def dump_pickle(self):
        """Make a Python pickle associated with the APBS input parameters.

        .. todo::  is this function still useful?
        """
        base_pqr_name = self.pqrpath.stem
        outname = base_pqr_name + "-input.p"
        with open(outname, "wb") as pfile:
            pickle.dump(self, pfile)


def split_input(filename):
    """Split the parallel input file into multiple async file names.

    :param filename:  the path to the original parallel input file
    :type filename:  str
    """
    nproc = 0
    with open(filename) as file_:
        text = ""
        while True:
            line = file_.readline()
            if not line:
                break
            text += line
            line = line.strip()
            if line.startswith("pdime"):  # Get # Procs
                words = line.split()
                nproc = int(words[1]) * int(words[2]) * int(words[3])
    if nproc == 0:
        errstr = f"{filename} is not a valid APBS parallel input file!\n"
        errstr = errstr + (
            "The inputgen script was unable to asynchronize this file!"
        )
        raise RuntimeError(errstr)
    base_pqr_name = Path(filename).stem
    for iproc in int(range(nproc)):
        outname = base_pqr_name + f"-PE{iproc}.in"
        outtext = text.replace("mg-para\n", f"mg-para\n    async {iproc}\n")
        with open(outname, "w") as outfile:
            outfile.write(outtext)


def build_parser():
    """Build argument parser."""
    desc = f"{TITLE_STR}\ninputgen: generating APBS input files since "
    desc += "(at least) 2004"
    parse = argparse.ArgumentParser(
        description=desc,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parse.add_argument(
        "--asynch",
        action="store_true",
        help="perform an asynchronous parallel calculation.",
    )
    parse.add_argument(
        "--split",
        action="store_true",
        help=(
            "split an existing parallel input file to multiple "
            "async input files."
        ),
    )
    parse.add_argument(
        "--potdx",
        action="store_true",
        help=("create an input to compute an electrostatic potential map."),
    )
    parse.add_argument(
        "--method",
        help=("force output file to write a specific APBS ELEC method."),
        choices=["para", "auto", "manual", "async"],
    )
    parse.add_argument(
        "--cfac",
        type=float,
        default=psize.CFAC,
        help=(
            "factor by which to expand molecular dimensions to "
            "get coarse grid dimensions."
        ),
    )
    parse.add_argument(
        "--fadd",
        type=float,
        default=psize.FADD,
        help=(
            "amount to add to molecular dimensions to get fine "
            "grid dimensions."
        ),
    )
    parse.add_argument(
        "--space",
        type=float,
        default=psize.SPACE,
        help="desired fine mesh resolution",
    )
    parse.add_argument(
        "--gmemfac",
        type=int,
        default=psize.GMEMFAC,
        help=(
            "number of bytes per grid point required for sequential "
            "MG calculation"
        ),
    )
    parse.add_argument(
        "--gmemceil",
        type=int,
        default=psize.GMEMCEIL,
        help=(
            "max MB allowed for sequential MG calculation; adjust "
            "this to force the script to perform faster calculations "
            "(which require more parallelism)"
        ),
    )
    parse.add_argument(
        "--ofrac",
        type=float,
        default=psize.OFRAC,
        help="overlap factor between mesh partitions (parallel)",
    )
    parse.add_argument(
        "--redfac",
        type=float,
        default=psize.REDFAC,
        help=(
            "the maximum factor by which a domain dimension can "
            "be reduced during focusing"
        ),
    )
    parse.add_argument(
        "--istrng", help="Ionic strength (M); Na+ and Cl- ions will be used"
    )
    parse.add_argument("filename")
    return parse


def main():
    """Main driver"""
    parser = build_parser()
    args = parser.parse_args()
    size = psize.Psize()
    filename = args.filename
    if args.split:
        split_input(filename)
    else:
        size.run_psize(filename)
        input_ = Input(
            filename,
            size,
            args.method,
            args.asynch,
            args.istrng,
            potdx=args.potdx,
        )
        path = Path(filename)
        output_path = f"{path.parent}/{path.stem}.in"
        input_.print_input_files(output_path)


if __name__ == "__main__":
    main()



"""Simple biomolecular structures.

This module contains the simpler structure objects used in PDB2PQR and their
associated methods.

.. codeauthor:: Todd Dolinsky
.. codeauthor:: Nathan Baker
"""

from typing_extensions import Self

from .config import BACKBONE
from .pdb import ATOM, HETATM


class Chain:
    """Chain class

    The chain class contains information about each chain within a given
    :class:`Biomolecule` object.
    """

    def __init__(self, chain_id: str):
        """Initialize the class.

        :param chain_id:  ID for this chain as denoted in the PDB
        :type chain_id:  str
        """
        self.chain_id = chain_id
        self.residues = []
        self.name = None

    def add_residue(self, residue):
        """Add a residue to the chain

        :param residue:  residue to be added
        :type residue:  Residue
        """
        self.residues.append(residue)

    def renumber_residues(self):
        """Renumber atoms.

        Renumber based on actual residue number and not PDB :makevar:`res_seq`
        """
        for count, residue in enumerate(self.residues, start=1):
            residue.set_res_seq(count)

    @property
    def atoms(self):
        """Return a list of Atom objects contained in this chain.

        :return: list of Atom objects
        :rtype: [Atom]
        """
        atomlist = []
        for residue in self.residues:
            my_list = residue.atoms
            for atom in my_list:
                atomlist.append(atom)
        return atomlist

    def __str__(self):
        output = [residue.letter_code() for residue in self.residues]
        return "".join(output)


class Atom:
    """Represent an atom.

    The Atom class inherits from the :class:`ATOM` object in :mod:`pdb`.
    This class used for adding fields not found in the PDB that may be useful
    for analysis.
    This class also simplifies code by combining :class:`ATOM` and
    :class:`HETATM` objects into a single class.
    """

    def __init__(
        self,
        atom: ATOM | HETATM | Self | None = None,
        type_="ATOM",
        residue=None,
    ):
        """Initialize the new Atom object by using the old object.

        :param atom:  the original ATOM object (could be None)
        :type atom:  ATOM
        :param type_:  either ATOM or HETATM
        :type type_:  str
        :param residue:  a pointer back to the parent residue object (could be
            None)
        :type residue:  Residue
        """
        self.type = None
        self.serial = None
        self.name = None
        self.alt_loc = None
        self.res_name = None
        self.chain_id = None
        self.res_seq = None
        self.ins_code = None
        self.x = None
        self.y = None
        self.z = None
        self.occupancy = None
        self.temp_factor = None
        self.seg_id = None
        self.element = None
        self.charge = None
        self.bonds = []
        self.reference = None
        self.residue = None
        self.radius = None
        self.ffcharge = None
        self.hdonor = 0
        self.hacceptor = 0
        self.cell = None
        self.added = 0
        self.optimizeable = 0
        self.refdistance = 0
        self.id = None
        self.mol2charge = None
        if type_ in ["ATOM", "HETATM"]:
            self.type = type_
        else:
            err = f"Invalid atom type {type_} (Atom Class IN structures.py)!"
            raise ValueError(err)
        if atom is not None:
            self.serial = atom.serial
            self.name = atom.name
            self.alt_loc = atom.alt_loc
            self.res_name = atom.res_name
            self.chain_id = atom.chain_id
            self.res_seq = atom.res_seq
            self.ins_code = atom.ins_code
            self.x = atom.x
            self.y = atom.y
            self.z = atom.z
            self.occupancy = atom.occupancy
            self.temp_factor = atom.temp_factor
            self.seg_id = atom.seg_id
            self.element = atom.element
            self.charge = atom.charge
            self.residue = residue
            if isinstance(atom, ATOM):
                # ATOM class doesn't have mol2charge
                self.mol2charge = None
            else:
                self.mol2charge = atom.mol2charge

    @classmethod
    def from_pqr_line(cls, line):
        """Create an atom from a PQR line.

        :param cls:  class for classmethod
        :type cls:  Atom
        :param line:  PQR line
        :type line:  str
        :returns:  new atom or None (for REMARK and similar lines)
        :rtype:  Atom
        :raises ValueError:  for problems parsing
        """
        atom = cls()
        words = [w.strip() for w in line.split()]
        token = words.pop(0)
        if token in [
            "REMARK",
            "TER",
            "END",
            "HEADER",
            "TITLE",
            "COMPND",
            "SOURCE",
            "KEYWDS",
            "EXPDTA",
            "AUTHOR",
            "REVDAT",
            "JRNL",
        ]:
            return None
        if token in ["ATOM", "HETATM"]:
            atom.type = token
        elif token[:4] == "ATOM":
            atom.type = "ATOM"
            words = [token[4:], *words]
        elif token[:6] == "HETATM":
            atom.type = "HETATM"
            words = [token[6:], *words]
        else:
            err = f"Unable to parse line: {line}"
            raise ValueError(err)
        atom.serial = int(words.pop(0))
        atom.name = words.pop(0)
        atom.res_name = words.pop(0)
        token = words.pop(0)
        try:
            atom.res_seq = int(token)
        except ValueError:
            atom.chain_id = token
            atom.res_seq = int(words.pop(0))
        token = words.pop(0)
        try:
            atom.x = float(token)
        except ValueError:
            atom.ins_code = token
            atom.x = float(words.pop(0))
        atom.y = float(words.pop(0))
        atom.z = float(words.pop(0))
        atom.charge = float(words.pop(0))
        atom.radius = float(words.pop(0))
        return atom

    @classmethod
    def from_qcd_line(cls, line, atom_serial):
        """Create an atom from a QCD (UHBD QCARD format) line.

        :param Atom cls:  class for classmethod
        :param str line:  PQR line
        :param int atom_serial:  atom serial number
        :returns:  new atom or None (for REMARK and similar lines)
        :rtype:  Atom
        :raises ValueError:  for problems parsing
        """
        atom = cls()
        words = [w.strip() for w in line.split()]
        token = words.pop(0)
        if token in [
            "REMARK",
            "TER",
            "END",
            "HEADER",
            "TITLE",
            "COMPND",
            "SOURCE",
            "KEYWDS",
            "EXPDTA",
            "AUTHOR",
            "REVDAT",
            "JRNL",
        ]:
            return None
        if token in ["ATOM", "HETATM"]:
            atom.type = token
        elif token[:4] == "ATOM":
            atom.type = "ATOM"
            words = [token[4:], *words]
        elif token[:6] == "HETATM":
            atom.type = "HETATM"
            words = [token[6:], *words]
        else:
            err = f"Unable to parse line: {line}"
            raise ValueError(err)
        atom.serial = int(atom_serial)
        atom.res_seq = int(words.pop(0))
        atom.res_name = words.pop(0)
        atom.name = words.pop(0)
        atom.x = float(words.pop(0))
        atom.y = float(words.pop(0))
        atom.z = float(words.pop(0))
        atom.charge = float(words.pop(0))
        atom.radius = float(words.pop(0))
        return atom

    def get_common_string_rep(self, chainflag=False):
        """Returns a string of the common column of the new atom type.

        Uses the :class:`ATOM` string output but changes the first field to
        either be ``ATOM`` or ``HETATM`` as necessary.
        This is used to create the output for PQR and PDB files.

        :return:  string with ATOM/HETATM field set appropriately
        :rtype:  str
        """
        outstr = ""
        tstr = self.type
        outstr += str.ljust(tstr, 6)[:6]
        tstr = f"{self.serial:d}"
        outstr += str.rjust(tstr, 5)[:5]
        outstr += " "
        tstr = self.name
        if len(tstr) == 4 or len(tstr.strip("FLIP")) == 4:
            outstr += str.ljust(tstr, 4)[:4]
        else:
            outstr += " " + str.ljust(tstr, 3)[:3]
        tstr = self.res_name
        if len(tstr) == 4:
            outstr += str.ljust(tstr, 4)[:4]
        else:
            outstr += " " + str.ljust(tstr, 3)[:3]
        outstr += " "
        tstr = self.chain_id if chainflag else ""
        outstr += str.ljust(tstr, 1)[:1]
        tstr = f"{self.res_seq:d}"
        outstr += str.rjust(tstr, 4)[:4]
        outstr += f"{self.ins_code}   " if self.ins_code != "" else "    "
        tstr = f"{self.x:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        tstr = f"{self.y:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        tstr = f"{self.z:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        return outstr

    def __str__(self):
        return self.get_pqr_string()

    def get_pqr_string(self, chainflag=False):
        """Returns a string of the atom type.

        Uses the :class:`ATOM` string output but changes the first field to
        either be ``ATOM`` or ``HETATM`` as necessary.
        This is used to create the output for PQR files.

        :return:  string with ATOM/HETATM field set appropriately
        :rtype:  str
        """
        outstr = self.get_common_string_rep(chainflag=chainflag)
        ffcharge = (
            f"{self.ffcharge:.4f}" if self.ffcharge is not None else "0.0000"
        )
        outstr += str.rjust(ffcharge, 8)[:8]
        ffradius = (
            f"{self.radius:.4f}" if self.radius is not None else "0.0000"
        )
        outstr += str.rjust(ffradius, 7)[:7]
        return outstr

    def get_pdb_string(self):
        """Returns a string of the atom type.

        Uses the :class:`ATOM` string output but changes the first field to
        either be ``ATOM`` or ``HETATM`` as necessary.
        This is for the PDB representation of the atom.
        The :mod:`propka` module depends on this being correct.

        :return:  string with ATOM/HETATM field set appropriately
        :rtype:  str
        """
        outstr = self.get_common_string_rep(chainflag=True)
        outstr += f"{self.occupancy:>6.2f}{self.temp_factor:>6.2f}      "
        outstr += f"{self.seg_id:4.4s}{self.element:>2.2s}{self.charge:2.2s}"
        return outstr

    @property
    def coords(self):
        """Return the x,y,z coordinates of the atom.

        .. todo:
           All atom coordinates should be converted to :mod:`numpy` arrays

        :return:  list of the coordinates
        :rtype:  [float, float, float]
        """
        return [self.x, self.y, self.z]

    def add_bond(self, bondedatom):
        """Add a bond to the list of bonds.

        :param bondedatom:  the atom to bond to
        :type bondedatom:  ATOM
        """
        self.bonds.append(bondedatom)

    @property
    def is_hydrogen(self):
        """Is this atom a Hydrogen atom?

        :return:  whether this atom is a hydrogen
        :rtype:  bool
        """
        return self.name[0] == "H"

    @property
    def is_backbone(self):
        """Return True if atom name is in backbone, otherwise False.

        :return:  whether atom is in backbone
        :rtype:  bool
        """
        return self.name in BACKBONE

    @property
    def has_reference(self):
        """Determine if the object has a reference object or not.

        All known atoms should have reference objects.

        :return:  whether atom has reference object
        :rtype:  bool
        """
        return self.reference is not None



"""Parser for topology XML files.

.. todo:
   There are some very silly classes in this module that should be combined
   or removed.

.. codeauthor::  Nathan Baker
.. codeauthor::  Yong Huang
"""

import logging
from xml import sax

_LOGGER = logging.getLogger(__name__)
#: File name of topology XML data
TOPOLOGYPATH = "TOPOLOGY.xml"


class TopologyHandler(sax.ContentHandler):
    """Handler for XML-based topology files.

    Assumes the following hierarchy of tags:

    * topology

      * residue

        * reference
        * titrationstate

          * tautomer

            * conformer

    .. todo:  Why isn't ``super().__init__`` called here?
    """

    def __init__(self):
        self.curr_element = None
        self.curr_atom = None
        self.curr_dihedral = None
        self.curr_reference = None
        self.curr_residue = None
        self.curr_titration_state = None
        self.curr_tautomer = None
        self.curr_conformer = None
        self.curr_conformer_add = None
        self.curr_conformer_remove = None
        self.residues = []
        self.incomplete = 0

    def startElement(self, tag_name, _):
        """Start element parsing.

        .. note:  overrides super-class method invocation

        :param tag_name:  name of XML element tag to start parsing
        :type tag_name:  str
        """
        if self.incomplete:
            return
        if tag_name == "topology":
            pass
        elif tag_name == "residue":
            if self.curr_residue is not None:
                _LOGGER.info("** Overwriting current topology_residue object!")
            self.curr_residue = TopologyResidue(self)
        elif tag_name == "reference":
            if self.curr_reference is not None:
                _LOGGER.info(
                    "** Overwriting current TopologyReference object!"
                )
            self.curr_reference = TopologyReference(self.curr_residue)
        elif tag_name == "titrationstate":
            if self.curr_titration_state is not None:
                _LOGGER.info(
                    "** Overwriting current topology_titration_state object!"
                )
            self.curr_titration_state = TopologyTitrationState(
                self.curr_residue
            )
        elif tag_name == "tautomer":
            if self.curr_tautomer is not None:
                _LOGGER.info("** Overwriting current Tautomer object!")
            self.curr_tautomer = TopologyTautomer(self.curr_titration_state)
        elif tag_name == "conformer":
            if self.curr_conformer is not None:
                _LOGGER.info("** Overwriting current Conformer object!")
            self.curr_conformer = TopologyConformer(self.curr_tautomer)
        elif tag_name == "name":
            self.curr_element = tag_name
        elif tag_name == "atom":
            if self.curr_conformer_add is not None:
                self.curr_atom = TopologyAtom(self.curr_conformer_add)
            elif self.curr_conformer_remove is not None:
                self.curr_atom = TopologyAtom(self.curr_conformer_remove)
            elif self.curr_reference is not None:
                self.curr_atom = TopologyAtom(self.curr_reference)
            else:
                _LOGGER.info("** Don't know what to do with this atom!")
        elif tag_name in ("x", "y", "z", "bond", "altname"):
            self.curr_element = tag_name
        elif tag_name == "dihedral":
            self.curr_element = tag_name
            if self.curr_conformer_add is not None:
                self.curr_dihedral = TopologyDihedral(self.curr_conformer_add)
            elif self.curr_conformer_remove is not None:
                self.curr_dihedral = TopologyDihedral(
                    self.curr_conformer_remove
                )
            elif self.curr_reference is not None:
                self.curr_dihedral = TopologyDihedral(self.curr_reference)
            else:
                _LOGGER.info("** Don't know what to do with this dihedral!")
        elif tag_name == "add":
            self.curr_conformer_add = TopologyConformerAdd(self.curr_conformer)
        elif tag_name == "remove":
            self.curr_conformer_remove = TopologyConformerRemove(
                self.curr_conformer
            )
        elif tag_name == "incomplete":
            self.incomplete = 1
        else:
            _LOGGER.info(f"** NOT handling {tag_name} start tag")

    def endElement(self, tag_name):
        """End parsing element.

        .. note: Overrides super-class method.

        :param tag_name:  name of XML element tag for which to end parsing
        :type tag_name:  str
        """
        if not self.incomplete:
            self.curr_element = None
            if tag_name in ("x", "y", "z", "name", "bond", "altname"):
                pass
            elif tag_name == "atom":
                self.curr_atom = None
            elif tag_name == "dihedral":
                self.curr_dihedral = None
            elif tag_name == "reference":
                self.curr_reference = None
            elif tag_name == "add":
                self.curr_conformer_add = None
            elif tag_name == "remove":
                self.curr_conformer_remove = None
            elif tag_name == "titrationstate":
                self.curr_titration_state = None
            elif tag_name == "conformer":
                self.curr_conformer = None
            elif tag_name == "tautomer":
                self.curr_tautomer = None
            elif tag_name == "residue":
                self.curr_residue = None
            elif tag_name != "topology":
                _LOGGER.info(f"** NOT handling {tag_name} end tag")
        elif tag_name == "incomplete":
            self.incomplete = 0

    def characters(self, text):
        """Parse characters in topology XML file.

        :param text:  XML character data
        :type text:  str
        """
        if text.isspace():
            return
        if not self.incomplete:
            if self.curr_element == "name":
                if self.curr_atom is not None:
                    self.curr_atom.name = text
                elif self.curr_conformer is not None:
                    self.curr_conformer.name = text
                elif self.curr_tautomer is not None:
                    self.curr_tautomer.name = text
                elif self.curr_titration_state is not None:
                    self.curr_titration_state.name = text
                elif self.curr_residue is not None:
                    self.curr_residue.name = text
                else:
                    _LOGGER.info(
                        f"    *** Don't know what to do with name {text}!"
                    )
            elif self.curr_element == "x":
                self.curr_atom.x = float(text)
            elif self.curr_element == "y":
                self.curr_atom.y = float(text)
            elif self.curr_element == "z":
                self.curr_atom.z = float(text)
            elif self.curr_element == "bond":
                self.curr_atom.bonds.append(text)
            elif self.curr_element == "altname":
                self.curr_atom.altname = text
            elif self.curr_element == "dihedral":
                self.curr_dihedral.atom_list = text
            else:
                _LOGGER.info(f"** NOT handling character text:  {text}")


class TopologyResidue:
    """A class for residue topology information.

    .. todo:
       This class (and other classes in this module) has (have) lots of
       repeated code that could be eliminated with better inheritance
       structures.
    """

    def __init__(self, topology_):
        """Initialize with a Topology object.

        :param topology_:  topology object
        """
        self.name = None
        self.reference = None
        self.titration_states = []
        self.topology = topology_
        self.topology.residues.append(self)

    def __str__(self):
        return self.name


class TopologyDihedral:
    """A class for dihedral angle topology information."""

    def __init__(self, parent):
        """Initialize with a parent that has a dihedral list.

        :param parent:  parent object with a dihedral list
        """
        self.parent = parent
        self.parent.dihedrals.append(self)
        self.atom_list = None

    def __str__(self):
        return self.atom_list


class TopologyAtom:
    """A class for atom topology information."""

    def __init__(self, parent):
        """Initialize with an upper-level class that contains an atom array.

        For example, initialize with :class:`TopologyReference` or
        :class:`TopologyConformerAddition`

        :param parent:  parent object with atom array
        """
        self.parent = parent
        self.parent.atoms.append(self)
        self.name = None
        self.x = None
        self.y = None
        self.z = None
        self.bonds = []
        self.altname = None

    def __str__(self):
        return self.name


class TopologyTitrationState:
    """A class for the titration state of a residue."""

    def __init__(self, topology_residue):
        """Initialize object.

        :param topology_residue:  residue topology
        :type topology_residue:  TopologyResidue
        """
        self.topology_residue = topology_residue
        self.topology_residue.titration_states.append(self)
        self.tautomers = []
        self.name = None

    def __str__(self):
        return self.name


class TopologyTautomer:
    """A class for topology tautomer information."""

    def __init__(self, topology_titration_state):
        """Initialize object.

        :param topology_titration_state:  titration state of this residue
        :type topology_titration_state:  TopologyTitrationState
        """
        self.topology_titration_state = topology_titration_state
        self.topology_titration_state.tautomers.append(self)
        self.conformers = []
        self.name = None

    def __str__(self):
        return self.name


class TopologyConformer:
    """A class for topology conformer information."""

    def __init__(self, topology_tautomer):
        """Initialize object.

        :param topology_tautomer:  tautomer for which to evaluate conformers
        :type topology_tautomer:  TopologyTautomer
        """
        self.topology_tautomer = topology_tautomer
        self.topology_tautomer.conformers.append(self)
        self.name = None
        self.conformer_adds = []
        self.conformer_removes = []

    def __str__(self):
        return self.name


class TopologyReference:
    """A class for the reference structure of a residue."""

    def __init__(self, topology_residue):
        """Initialize object.

        :param topology_residue:  reference residue
        :type topology_residue:  TopologyResidue
        """
        self.topology_residue = topology_residue
        self.topology_residue.reference = self
        self.atoms = []
        self.dihedrals = []


class TopologyConformerAdd:
    """A class for adding atoms to a conformer."""

    def __init__(self, topology_conformer):
        """Initialize object.

        :param topology_conformer:  conformer to which to add atoms
        :type topology_conformer:  TopologyConformer
        """
        self.topology_conformer = topology_conformer
        self.topology_conformer.conformer_adds.append(self)
        self.atoms = []
        self.name = None
        self.dihedrals = []


class TopologyConformerRemove:
    """A class for removing atoms from a conformer."""

    def __init__(self, topology_conformer):
        """Initialize object.

        :param topology_conformer:  conformer to which to add atoms
        :type topology_conformer:  TopologyConformer
        """
        self.topology_conformer = topology_conformer
        self.topology_conformer.conformer_removes.append(self)
        self.atoms = []
        self.name = None


class Topology:
    """Contains the structured definitions of residue reference coordinates
    as well as alternate titration, conformer, and tautomer states.
    """

    def __init__(self, topology_file):
        """Initialize object.

        :param topology_file:  topology file object ready for reading
        :type topology_file:  file-like object
        """
        handler = TopologyHandler()
        sax.make_parser()
        sax.parseString(topology_file.read(), handler)
        self.residues = handler.residues



"""Ligand support functions.

.. todo::
   Some of the definitions in this module belong in a configuration file other
   than here.

.. codeauthor:: Jens Erik Nielsen
"""

#: Radii for different atom types.
#: When using these tables, the most specific Sybyl atom type should be used
#: first and then the generic element should be used
RADII = {
    # NOTE - these are not the original PARSE radii but they are the ones
    # included in the previous version of PDB2PKA so I'm preserving them for
    # posterity. There's a claim they came from
    # http://amber.ch.ic.ac.uk/archive/ but that link no longer works.
    "not parse - do not use": {
        "C": 1.70,
        "N": 1.50,
        "O": 1.40,
        "S": 1.85,
        "H": 1.00,
        "Br": 2.50,
        "F": 1.20,
        "P": 1.90,
        "Cl": 1.75,
    },
    #: These are the PARSE radii from Table 4 of
    #: http://doi.org/10.1021/j100058a043
    "parse": {
        "C.1": 2.00,
        "C.2": 2.00,
        "C.3": 2.00,
        "C": 1.70,
        "H": 1.00,
        "O": 1.40,
        "N": 1.50,
        "S": 1.85,
    },
    #: These are the ZAP radii from Table 2 of
    #: http://doi.org/10.1021/jm070549%2B. Bondi radii should be used for
    #: atoms not found in this table.
    "zap9": {
        "C": 1.87,
        "H": 1.10,
        "O.co2": 1.76,
        "N": 1.40,
        "S": 2.15,
        "F": 2.40,
        "Cl": 1.82,
        "I": 2.65,
    },
    #: These are the Bondi radii from Table 2 of
    #: http://doi.org/10.1021/jm070549%2B
    "bondi-zap": {
        "C": 1.7,
        "H": 1.20,
        "O.co2": 1.52,
        "N": 1.55,
        "S": 1.80,
        "F": 1.47,
        "Cl": 1.75,
        "I": 1.98,
    },
    #: These are the Bondi radii from Table I of
    #: http://doi.org/10.1021/j100785a001. NOTE - there are some variations to
    #: the halogens in Table V that we might want to consider in the future.
    "bondi": {
        "H": 1.20,
        "He": 1.40,
        "C": 1.70,
        "N": 1.55,
        "O": 1.52,
        "F": 1.47,
        "Ne": 1.54,
        "Si": 2.10,
        "P": 1.80,
        "S": 1.80,
        "Cl": 1.75,
        "Ar": 1.88,
        "As": 1.85,
        "Se": 1.90,
        "Br": 1.85,
        "Kr": 2.02,
        "Te": 2.06,
        "I": 1.98,
        "Xe": 2.16,
    },
}


#: Numbers of valence electrons for the groups of the periodic table
VALENCE_BY_GROUP = {1: 1, 2: 2, 13: 3, 14: 4, 15: 5, 16: 6, 17: 7, 18: 8}


#: Groups of the periodic table
ELEMENT_BY_GROUP = {
    1: ["H", "Li", "Na", "K", "Rb", "Cs", "Fr"],
    2: ["Be", "Mg", "Ca", "Sr", "Ba", "Ra"],
    13: ["B", "Al", "Ga", "In", "Tl", "Nh"],
    14: ["C", "Si", "Ge", "Sn", "Pb", "Fl"],
    15: ["N", "P", "As", "Sb", "Bi", "Mc"],
    16: ["O", "S", "Se", "Te", "Po", "Lv"],
    17: ["F", "Cl", "Br", "I", "At", "Ts"],
    18: ["He", "Ne", "Ar", "Kr", "Xe", "Rn", "Og"],
}


#: Valence electrons by element
VALENCE_BY_ELEMENT = {}
for group, elem_list in ELEMENT_BY_GROUP.items():
    for elem in elem_list:
        VALENCE_BY_ELEMENT[elem] = VALENCE_BY_GROUP[group]


#: Numbers of non-bonded electrons for Sybyl-type atoms.  Adapted from
#: https://onlinelibrary.wiley.com/doi/abs/10.1002/jcc.540100804 (Table I).
NONBONDED_BY_TYPE = {
    "Al": 0,
    "Br": 6,
    "C.1": 0,
    "C.2": 0,
    "C.3": 0,
    "C.ar": 0,
    "Ca": 0,
    "Cl": 6,
    "F": 6,
    "H": 0,
    "I": 6,
    "K": 0,
    "Li": 0,
    "N.1": 2,
    "N.2": 2,
    "N.3": 2,
    "N.4": 0,
    "N.am": 0,
    "N.ar": 2,
    "N.pl3": 0,
    "Na": 0,
    "O.2": 4,
    "O.3": 4,
    "P.3": 0,
    "S.2": 4,
    "S.3": 4,
    "S.o": 2,
    "S.o2": 0,
    "Si": 0,
    "O.co2": 4.5,
}



#!/usr/bin/python
"""Get dimensions and other information from a PQR file.

.. todo:: This code could be combined with :mod:`inputgen`.

.. todo:: This code should be moved to the APBS code base.

.. codeauthor:: Dave Sept
.. codeauthor:: Nathan Baker
.. codeauthor:: Todd Dolinksy
.. codeauthor:: Yong Huang
"""

import argparse
import logging
from math import log

from .config import TITLE_STR

#: The number of Angstroms added to the molecular dimensions to determine the
#: find grid dimensions
FADD = 20.0
#: The fine grid dimensions are multiplied by this constant to calculate the
#: coarse grid dimensions
CFAC = 1.7
#: Desired fine grid spacing (in Angstroms)
SPACE = 0.50
#: Approximate memory usage (in bytes) can be estimated by multiplying the
#: number of grid points by this constant
GMEMFAC = 200
#: Maxmimum memory (in MB) to be used for a calculation
GMEMCEIL = 400
#: The fractional overlap between grid partitions in a parallel focusing
#: calculation
OFRAC = 0.1
#: The maximum factor by which a domain can be "shrunk" during a focusing
#: calculation
REDFAC = 0.25
_LOGGER = logging.getLogger(__name__)


class Psize:
    """Class for parsing input files and suggesting settings."""

    def __init__(
        self,
        cfac=CFAC,
        fadd=FADD,
        space=SPACE,
        gmemfac=GMEMFAC,
        gmemceil=GMEMCEIL,
        ofrac=OFRAC,
        redfac=REDFAC,
    ):
        """Initialize.

        :param cfac:  factor by which to expand molecular dimensions to get
            coarse grid dimensions
        :type cfac:  float
        :param fadd:  amount (in Angstroms) to add to molecular dimensions to
            get the fine grid dimensions
        :type fadd:  float
        :param space:  desired fine mesh resolution (in Angstroms)
        :type space:  float
        :param gmemfac:  number of bytes per grid point required for
            a sequential multigrid calculation
        :type gmemfac:  float
        :param gmemceil:  maximum memory (in MB) allowed for sequential
            multigrid calculation. Adjust this value to force the script to
            perform faster calculations (which require more parallelism).
        :type gmemceil:  float
        :param ofrac:  overlap factor between mesh partitions in parallel
            focusing calculation
        :type ofrac:  float
        :param redfac:  the maximum factor by which a domain dimension can be
            reduced during focusing
        :type redfac:  float
        """
        self.minlen = [None, None, None]
        self.maxlen = [None, None, None]
        self.cfac = cfac
        self.fadd = fadd
        self.space = space
        self.gmemfac = gmemfac
        self.gmemceil = gmemceil
        self.ofrac = ofrac
        self.redfac = redfac
        self.charge = 0.0
        self.gotatom = 0
        self.gothet = 0
        self.mol_length = [0.0, 0.0, 0.0]
        self.center = [0.0, 0.0, 0.0]
        self.coarse_length = [0.0, 0.0, 0.0]
        self.fine_length = [0.0, 0.0, 0.0]
        self.ngrid = [0, 0, 0]
        self.proc_grid = [0.0, 0.0, 0.0]
        self.nsmall = [0, 0, 0]
        self.nfocus = 0

    def parse_string(self, structure):
        """Parse the input structure as a string in PDB or PQR format.

        :param structure:  input structure as string in PDB or PQR format.
        :type structure:  str
        """
        lines = structure.split("\n")
        self.parse_lines(lines)

    def parse_input(self, filename):
        """Parse input structure file in PDB or PQR format.

        :param filename:  string with path to PDB- or PQR-format file.
        :type filename:  str
        """
        with open(filename, encoding="utf-8") as file_:
            self.parse_lines(file_.readlines())

    def parse_lines(self, lines):
        """Parse the PQR/PDB lines.

        .. todo::
           This is messed up. Why are we parsing the PQR manually here when
           we already have other routines to do that?  This function should
           be replaced by a call to existing routines.

        :param lines:  PDB/PQR lines to parse
        :type lines:  [str]
        """
        for line in lines:
            if line.find("ATOM") == 0:
                self.gotatom += 1
            elif line.find("HETATM") == 0:
                self.gothet = self.gothet + 1
            subline = line[30:].replace("-", " -")
            words = subline.split()
            if len(words) < 5:
                continue
            self.charge = self.charge + float(words[3])
            rad = float(words[4])
            center = [float(word) for word in words[0:3]]
            for i in range(3):
                if self.minlen[i] is None or center[i] - rad < self.minlen[i]:
                    self.minlen[i] = center[i] - rad
                if self.maxlen[i] is None or center[i] + rad > self.maxlen[i]:
                    self.maxlen[i] = center[i] + rad

    def set_length(self, maxlen, minlen):
        """Compute molecular dimensions, adjusting for zero-length values.

        .. todo:: Replace hard-coded values in this function.

        :param maxlen:  maximum dimensions from molecule
        :type maxlen:  [float, float, float]
        :param minlen:  minimum dimensions from molecule
        :type minlen:  [float, float, float]
        :return:  molecular dimensions
        :rtype:  [float, float, float]
        """
        for i in range(3):
            self.mol_length[i] = maxlen[i] - minlen[i]
            self.mol_length[i] = max(self.mol_length[i], 0.1)
        return self.mol_length

    def set_coarse_grid_dims(self, mol_length):
        """Compute coarse mesh lengths.

        :param mol_length:  input molecule lengths
        :type mol_length:  [float, float, float]
        :return:  coarse grid dimensions
        :rtype:  [float, float, float]
        """
        for i in range(3):
            self.coarse_length[i] = self.cfac * mol_length[i]
        return self.coarse_length

    def set_fine_grid_dims(self, mol_length, coarse_length):
        """Compute fine mesh lengths.

        :param mol_length:  input molecule lengths
        :type mol_length:  [float, float, float]
        :param coarse_length:  coarse grid lengths
        :type coarse_length:  [float, float, float]
        :return:  fine grid lengths
        :rtype:  [float, float, float]
        """
        for i in range(3):
            self.fine_length[i] = mol_length[i] + self.fadd
            self.fine_length[i] = min(self.fine_length[i], coarse_length[i])
        return self.fine_length

    def set_center(self, maxlen, minlen):
        """Compute molecular center.

        :param maxlen:  maximum molecule lengths
        :type maxlen:  [float, float, float]
        :param minlen:  minimum molecule lengths
        :type minlen:  [float, float, float]
        :return:  center of molecule
        :rtype:  [float, float, float]
        """
        for i in range(3):
            self.center[i] = (maxlen[i] + minlen[i]) / 2
        return self.center

    def set_fine_grid_points(self, fine_length):
        """Compute mesh grid points, assuming 4 levels in multigrid hierarchy.

        .. todo:: remove hard-coded values from this function.

        :param fine_length:  lengths of the fine grid
        :type fine_length:  [float, float, float]
        :return:  number of grid points in each direction
        :rtype:  [int, int, int]
        """
        temp_num = [0, 0, 0]
        for i in range(3):
            temp_num[i] = int(fine_length[i] / self.space + 0.5)
            self.ngrid[i] = 32 * (int((temp_num[i] - 1) / 32.0 + 0.5)) + 1
            self.ngrid[i] = max(self.ngrid[i], 33)
        return self.ngrid

    def set_smallest(self, ngrid):
        """Set smallest dimensions.

        Compute parallel division of domain in case the memory requirements
        for the calculation are above the memory ceiling. Find the smallest
        dimension and see if the number of grid points in that dimension will
        fit below the memory ceiling Reduce nsmall until an nsmall^3 domain
        will fit into memory.

        .. todo:: Remove hard-coded values from this function.

        :param ngrid:  number of grid points
        :type ngrid:  [int, int, int]
        :return:  smallest number of grid points in each direction to fit in
            memory
        :rtype:  [int, int, int]
        """
        nsmall = [ngrid[i] for i in range(3)]
        while 1:
            nsmem = 200.0 * nsmall[0] * nsmall[1] * nsmall[2] / 1024 / 1024
            if nsmem < self.gmemceil:
                break
            else:
                i = nsmall.index(max(nsmall))
                nsmall[i] = 32 * ((nsmall[i] - 1) / 32 - 1) + 1
                if nsmall[i] <= 0:
                    _LOGGER.error(
                        "You picked a memory ceiling that is too small"
                    )
                    raise ValueError(nsmall[i])
        self.nsmall = nsmall
        return nsmall

    def set_proc_grid(self, ngrid, nsmall):
        """Calculate the number of processors required in a parallel focusing
        calculation to span each dimension of the grid given the grid size
        suitable for memory constraints.

        :param ngrid:  number of needed grid points
        :type ngrid:  [int, int, int]
        :param nsmall:  number of grid points that will fit in memory
        :type nsmall:  [int, int, int]
        :return:  number of processors needed in each direction
        :rtype:  [int, int, int]
        """
        zofac = 1 + 2 * self.ofrac
        for i in range(3):
            self.proc_grid[i] = ngrid[i] / float(nsmall[i])
            if self.proc_grid[i] > 1:
                self.proc_grid[i] = int(zofac * ngrid[i] / nsmall[i] + 1.0)
        return self.proc_grid

    def set_focus(self, fine_length, nproc, coarse_length):
        """Calculate the number of levels of focusing required for each processor subdomain.

        :param fine_length:  fine grid length
        :type fine_length:  [float, float, float]
        :param nproc:  number of processors in each dimension
        :type nproc:  [int, int, int]
        :param coarse_length:  coarse grid length
        :type coarse_length:  [float, float, float]
        """
        nfoc = [0, 0, 0]
        for i in range(3):
            nfoc[i] = int(
                log((fine_length[i] / nproc[i]) / coarse_length[i])
                / log(self.redfac)
                + 1.0
            )
        nfocus = nfoc[0]
        nfocus = max(nfoc[1], nfocus)
        nfocus = max(nfoc[2], nfocus)
        if nfocus > 0:
            nfocus = nfocus + 1
        self.nfocus = nfocus

    def set_all(self):
        """Set up all of the things calculated individually above."""
        maxlen = self.maxlen
        minlen = self.minlen
        self.set_length(maxlen, minlen)
        mol_length = self.mol_length
        self.set_coarse_grid_dims(mol_length)
        coarse_length = self.coarse_length
        self.set_fine_grid_dims(mol_length, coarse_length)
        fine_length = self.fine_length
        self.set_center(maxlen, minlen)
        self.set_fine_grid_points(fine_length)
        ngrid = self.ngrid
        self.set_smallest(ngrid)
        nsmall = self.nsmall
        self.set_proc_grid(ngrid, nsmall)
        nproc = self.proc_grid
        self.set_focus(fine_length, nproc, coarse_length)

    def run_psize(self, filename):
        """Parse input PQR file and set parameters.

        :param filename:  path of PQR file
        :type filename:  str
        """
        self.parse_input(filename)
        self.set_all()

    def __str__(self):
        """Return a string with the formatted results.

        :return:  string with formatted results
        :rtype:  str
        """
        str_ = "\n"
        if self.gotatom > 0:
            maxlen = self.maxlen
            minlen = self.minlen
            charge = self.charge
            mol_length = self.mol_length
            coarse_length = self.coarse_length
            fine_length = self.fine_length
            center = self.center
            ngrid = self.ngrid
            nsmall = self.nsmall
            nproc = self.proc_grid
            nfocus = self.nfocus
            # Compute memory requirements
            nsmem = 200.0 * nsmall[0] * nsmall[1] * nsmall[2] / 1024 / 1024
            gmem = 200.0 * ngrid[0] * ngrid[1] * ngrid[2] / 1024 / 1024
            # Print the calculated entries
            str_ += "######## MOLECULE INFO ########\n"
            str_ += f"Number of ATOM entries = {self.gotatom}\n"
            str_ += f"Number of HETATM entries (ignored) = {self.gothet}\n"
            str_ += f"Total charge = {charge:.3f} e\n"
            str_ += f"Dimensions = {mol_length[0]:.3f} Å x "
            str_ += f"{mol_length[1]:.3f} Å x {mol_length[2]:.3f} Å\n"
            str_ += f"Center = {center[0]:.3f} Å x {center[1]:.3f} Å x "
            str_ += f"{center[2]:.3f} Å\n"
            str_ += f"Lower corner = {minlen[0]:.3f} Å x {minlen[1]:.3f} Å x "
            str_ += f"{minlen[2]:.3f} Å\n"
            str_ += f"Upper corner = {maxlen[0]:.3f} Å x {maxlen[1]:.3f} Å x "
            str_ += f"{maxlen[2]:.3f} Å\n"
            str_ += "\n"
            str_ += "######## GENERAL CALCULATION INFO ########\n"
            str_ += f"Course grid dims = {coarse_length[0]:.3f} Å x "
            str_ += f"{coarse_length[1]:.3f} Å x "
            str_ += f"{coarse_length[2]:.3f} Å\n"
            str_ += f"Fine grid dims = {fine_length[0]:.3f} Å x "
            str_ += f"{fine_length[1]:.3f} Å x "
            str_ += f"{fine_length[2]:.3f} Å\n"
            str_ += f"Num. fine grid pts. = {ngrid[0]:d} Å x "
            str_ += f"{ngrid[1]:d} Å x "
            str_ += f"{ngrid[2]:d} Å\n"
            str_ += "\n"
            if gmem > self.gmemceil:
                str_ += f"Parallel solve required ({gmem:.3f} MB > "
                str_ += f"{self.gmemceil:.3f} MB)\n"
                str_ += "Total processors required = "
                str_ += f"{nproc[0] * nproc[1] * nproc[2]}\n"
                str_ += f"Proc. grid = {nproc[0]:d} x {nproc[1]:d} x "
                str_ += f"{nproc[2]:d}\n"
                str_ += f"Grid pts. on each proc. = {nsmall[0]:d} x "
                str_ += f"{nsmall[1]:d} x {nsmall[2]:d}\n"
                xglob = nproc[0] * round(
                    nsmall[0] / (1 + 2 * self.ofrac - 0.001)
                )
                yglob = nproc[1] * round(
                    nsmall[1] / (1 + 2 * self.ofrac - 0.001)
                )
                zglob = nproc[2] * round(
                    nsmall[2] / (1 + 2 * self.ofrac - 0.001)
                )
                if nproc[0] == 1:
                    xglob = nsmall[0]
                if nproc[1] == 1:
                    yglob = nsmall[1]
                if nproc[2] == 1:
                    zglob = nsmall[2]
                str_ += "Fine mesh spacing = "
                str_ += f"{fine_length[0] / (xglob - 1):g} x "
                str_ += f"{fine_length[1] / (yglob - 1):g} x "
                str_ += f"{fine_length[2] / (zglob - 1):g} A\n"
                str_ += "Estimated mem. required for parallel solve = "
                str_ += f"{nsmem:.3f} MB/proc.\n"
                ntot = nsmall[0] * nsmall[1] * nsmall[2]
            else:
                str_ += "Fine mesh spacing = "
                str_ += f"{fine_length[0] / (ngrid[0] - 1):g} x "
                str_ += f"{fine_length[1] / (ngrid[1] - 1):g} x "
                str_ += f"{fine_length[2] / (ngrid[2] - 1):g} A\n"
                str_ += "Estimated mem. required for sequential solve = "
                str_ += f"{gmem:.3f} MB\n"
                ntot = ngrid[0] * ngrid[1] * ngrid[2]
            str_ += f"Number of focusing operations = {nfocus}\n"
            str_ += "\n"
            str_ += "######## ESTIMATED REQUIREMENTS ########\n"
            str_ += "Memory per processor = "
            str_ += f"{200.0 * ntot / 1024 / 1024:.3f} MB\n"
            str_ += "Grid storage requirements (ASCII) = "
            grid_storage_req = (
                8.0 * 12 * nproc[0] * nproc[1] * nproc[2] * ntot / 1024 / 1024
            )
            str_ += f"{grid_storage_req:.3f} MB\n"
            str_ += "\n"
        else:
            str_ = "No ATOM entries in file!\n\n"
        return str_


def build_parser():
    """Build argument parser.

    :return:  argument parser
    :rtype:  argparse.ArgumentParser
    """
    desc = f"{TITLE_STR}\npsize: figuring out the size of electrostatics "
    desc += "calculations since (at least) 2002."

    parser = argparse.ArgumentParser(
        description=desc,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--cfac",
        default=CFAC,
        type=float,
        help=(
            "Factor by which to expand molecular dimensions to "
            "get coarse grid dimensions"
        ),
    )
    parser.add_argument(
        "--fadd",
        default=FADD,
        type=float,
        help="Amount to add to mol dims to get fine grid dims",
    )
    parser.add_argument(
        "--space",
        default=SPACE,
        type=float,
        help="Desired fine mesh resolution",
    )
    parser.add_argument(
        "--gmemfac",
        default=GMEMFAC,
        type=int,
        help=(
            "Number of bytes per grid point required for "
            "sequential MG calculation"
        ),
    )
    parser.add_argument(
        "--gmemceil",
        default=GMEMCEIL,
        type=int,
        help=(
            "Max MB allowed for sequential MG calculation. "
            "Adjust this to force the script to perform faster "
            "calculations (which require more parallelism)."
        ),
    )
    parser.add_argument(
        "--ofrac",
        default=OFRAC,
        type=float,
        help="Overlap factor between mesh partitions",
    )
    parser.add_argument(
        "--redfac",
        default=REDFAC,
        type=float,
        help=(
            "The maximum factor by which a domain dimension "
            "can be reduced during focusing"
        ),
    )
    parser.add_argument("mol_path", help="Path to PQR file.")
    return parser
